import bson.objectid

from cgi_parsing import *
from query_execution import mongo_result_list


################################################################################

def parse_variant_parameters(byc):

    if not "variant_definitions" in byc:
        return

    variant_pars = { }
    v_p_defs = byc["variant_definitions"]["parameters"]
    v_t_als = byc["variant_definitions"]["variant_state_aliases"]
    v_t_defs = byc["variant_definitions"]["variant_states"]

    for p_k in v_p_defs.keys():
        v_default = None
        if "default" in v_p_defs[ p_k ]:
            v_default = v_p_defs[ p_k ][ "default" ]
        variant_pars[ p_k ] = v_default
        if p_k in byc["form_data"]:
            variant_pars[ p_k ] = byc["form_data"][p_k]

        if variant_pars[ p_k ] is None:
            variant_pars.pop(p_k)

    # value checks
    v_p_c = { }
    translate_reference_name(variant_pars, byc)

    for p_k in variant_pars.keys():
        if not p_k in v_p_defs.keys():
            continue
        v_p = variant_pars[ p_k ]
        if "variant_type" in p_k:
            if v_p in v_t_als:
                v_t_k = v_t_als[v_p]
                if "LiteralSequenceExpression" in v_t_k:
                    v_p_c[ p_k ] = None
                else:
                    v_p_c[ p_k ] = { "$in": v_t_defs[v_t_k]["child_terms"] }
        elif "array" in v_p_defs[ p_k ]["type"]:
            v_l = set()
            for v in v_p:
                if re.compile( v_p_defs[ p_k ][ "items" ][ "pattern" ] ).match( str( v ) ):
                    if "integer" in v_p_defs[ p_k ][ "items" ][ "type" ]:
                        v = int( v )
                    v_l.add( v )
            v_p_c[ p_k ] = sorted( list(v_l) )
        else:
            if re.compile( v_p_defs[ p_k ][ "pattern" ] ).match( str( v_p ) ):
                if "integer" in v_p_defs[ p_k ][ "type" ]:
                    v_p = int( v_p )
                v_p_c[ p_k ] = v_p

    byc.update( { "variant_pars": v_p_c } )

################################################################################

def translate_reference_name(variant_pars, byc):

    if not "reference_name" in variant_pars:
        return variant_pars

    r_n = variant_pars[ "reference_name" ]
    r_a = byc["variant_definitions"]["refseq_aliases"]

    if not r_n in r_a.keys():
        variant_pars.pop("reference_name")
        return variant_pars

    variant_pars.update({"reference_name": r_a[r_n] })

    return variant_pars

################################################################################

def get_variant_request_type(byc):

    """podmd
    This method guesses the type of variant request, based on the complete
    fulfillment of the required parameters (all of `all_of`, one if `one_of`).
    In case of multiple types the one with most matched parameters is prefered.
    This may be changed to using a pre-defined request type and using this as
    completeness check only.
    TODO: Verify by schema ...
    podmd"""

    if not "variant_pars" in byc:
        return

    variant_request_type = "no correct variant request"

    v_pars = byc["variant_pars"]
    v_p_defs = byc["variant_definitions"]["parameters"]

    brts = byc["variant_definitions"]["request_types"]
    brts_k = brts.keys()
    
    # DODO HACK: setting to range request if start and end with one value
    if "start" in v_pars and "end" in v_pars:
        if len(v_pars[ "start" ]) == 1:
            if len(v_pars[ "end" ]) == 1:
                brts_k = [ "variantRangeRequest" ]

    vrt_matches = [ ]

    for vrt in brts_k:

        matched_par_no = 0
        needed_par_no = 0
        if "one_of" in brts[vrt]:
            needed_par_no = 1
            for one_of in brts[vrt][ "one_of" ]:
                if one_of in v_pars:
                    matched_par_no = 1
                    continue
        
        if "all_of" in brts[vrt]:
            needed_par_no += len( brts[vrt][ "all_of" ] )

            for required in brts[vrt][ "all_of" ]:
                if required in v_pars:
                    matched_par_no += 1
        
        # print("{} {} of {}".format(vrt, matched_par_no, needed_par_no))

        if matched_par_no >= needed_par_no:
            vrt_matches.append( { "type": vrt, "par_no": matched_par_no } )


    if len(vrt_matches) > 0:
        vrt_matches = sorted(vrt_matches, key=lambda k: k['par_no'], reverse=True)
        variant_request_type = vrt_matches[0]["type"]

    byc.update( { "variant_request_type": variant_request_type } )

################################################################################

def create_variantTypeRequest_query( byc ):

    if byc["variant_request_type"] != "variantTypeRequest":
        return

    """
    A query just for a variant type w/o any additional parameter has to be blocked
    due to size concerns.
    Queries with additional parameters (start ...) will be handled through other
    query types.
    """

    if len(byc.get("filters", [])) < 1:
        return

    vp = byc["variant_pars"]
    v_p_defs = byc["variant_definitions"]["parameters"]

    v_q_l = [
        { v_p_defs["variant_type"]["db_key"]: vp[ "variant_type" ] }
    ]

    if "reference_name" in vp:
        v_q_l.append( { v_p_defs["reference_name"]["db_key"]: vp[ "reference_name" ] })

    if len(v_q_l) == 1:
        v_q = v_q_l[0]
    else:     
        v_q = { "$and": v_q_l }

    expand_variant_query(v_q, byc)

################################################################################

def create_variantIdRequest_query( byc ):

    if byc["variant_request_type"] != "variantIdRequest":
        return

    # query database for gene and use coordinates to create range query
    vp = byc["variant_pars"]
    v_p_defs = byc["variant_definitions"]["parameters"]

    if "_id" in vp:
        v_q = {v_p_defs["_id"]["db_key"] : bson.objectid.ObjectId(vp["_id"])}
    elif "id" in vp:
        v_q = { 
            "$or": [
                {v_p_defs["_id"]["db_key"] : bson.objectid.ObjectId(vp["id"])},
                { v_p_defs["id"]["db_key"] : vp[ "id" ] }
            ]
        }
    else:
        return

    expand_variant_query(v_q, byc)

################################################################################

def create_geneVariantRequest_query( byc ):

    if byc["variant_request_type"] != "geneVariantRequest":
        return

    # query database for gene and use coordinates to create range query
    vp = byc["variant_pars"]
    v_p_defs = byc["variant_definitions"]["parameters"]


    query = { "symbol" : vp[ "gene_id" ] }

    results, error = mongo_result_list( "progenetix", "genes", query, { '_id': False } )

    # Since this is a pre-processor to the range request
    byc["variant_pars"].update( {
        "reference_name": "refseq:{}".format(results[0]["accession_version"]),
        "start": [ results[0]["start"] ],
        "end": [ results[0]["end"] ]
    } )

    # translate_reference_name(byc["variant_pars"], byc)
    byc.update( {"variant_request_type": "variantRangeRequest"} )
    create_variantRangeRequest_query( byc )

################################################################################

def create_variantAlleleRequest_query( byc ):

    """podmd
 
    podmd"""

    if byc["variant_request_type"] != "variantAlleleRequest":
        return

    vp = byc["variant_pars"]
    v_p_defs = byc["variant_definitions"]["parameters"]

    # TODO: Regexes for ref or alt with wildcard characters

    v_q_l = [
        { v_p_defs["reference_name"]["db_key"]: vp[ "reference_name" ] },
        { v_p_defs["start"]["db_key"]: int(vp[ "start" ][0]) }
    ]
    for p in [ "reference_bases", "alternate_bases" ]:
        if not vp[ p ] == "N":
            if "N" in vp[ p ]:
                rb = vp[ p ].replace("N", ".")
                v_q_l.append( { v_p_defs[p]["db_key"]: { '$regex': rb } } )
            else:
                 v_q_l.append( { v_p_defs[p]["db_key"]: vp[ p ] } )
        
    v_q = { "$and": v_q_l }

    expand_variant_query(v_q, byc)

################################################################################

def create_variantCNVrequest_query( byc ):

    if not byc["variant_request_type"] in [ "variantCNVrequest" ]:
        return

    vp = byc["variant_pars"]
    v_p_defs = byc["variant_definitions"]["parameters"]

    v_q = { "$and": [
        { v_p_defs["reference_name"]["db_key"]: vp[ "reference_name" ] },
        { v_p_defs["start"]["db_key"]: { "$lt": vp[ "start" ][-1] } },
        { v_p_defs["end"]["db_key"]: { "$gte": vp[ "end" ][0] } },
        { v_p_defs["start"]["db_key"]: { "$gte": vp[ "start" ][0] } },
        { v_p_defs["end"]["db_key"]: { "$lt": vp[ "end" ][-1] } },
        create_in_query_for_parameter("variant_type", v_p_defs["variant_type"]["db_key"], vp)
    ]}

    expand_variant_query(v_q, byc)

################################################################################

def create_variantRangeRequest_query( byc ):

    if not byc["variant_request_type"] in [ "variantRangeRequest" ]:
        return
    
    vp = byc["variant_pars"]
    v_p_defs = byc["variant_definitions"]["parameters"]

    v_q_l = [
        { v_p_defs["reference_name"]["db_key"]: vp[ "reference_name" ] },
        { v_p_defs["start"]["db_key"]: { "$lt": int(vp[ "end" ][-1]) } },
        { v_p_defs["end"]["db_key"]: { "$gt": int(vp[ "start" ][0]) } }
    ]

    p_n = "variant_min_length"
    if p_n in vp:
        v_q_l.append( { v_p_defs[p_n]["db_key"]: { "$gte" : vp[p_n] } } )
    p_n = "variant_max_length"
    if "variant_max_length" in vp:
        v_q_l.append( { v_p_defs[p_n]["db_key"]: { "$lte" : vp[p_n] } } )

    p_n = "variant_type"
    if p_n in vp:
        v_q_l.append( create_in_query_for_parameter(p_n, v_p_defs[p_n]["db_key"], vp) )
    elif "alternate_bases" in vp:
        # the N wildcard stands for any length alt bases so can be ignored
        if vp[ "alternate_bases" ] == "N":
             v_q_l.append( { v_p_defs["alternate_bases"]["db_key"]: {'$regex': "." } } )
        else:
            v_q_l.append( { v_p_defs["alternate_bases"]["db_key"]: vp[ "alternate_bases" ] } )

    v_q = { "$and": v_q_l }

    expand_variant_query(v_q, byc)

################################################################################

def expand_variant_query(variant_query, byc):

    if "variants" in byc["queries"]:
        byc["queries"].update({"variants": { "$and": [ byc["queries"]["variants"], variant_query ] } } )
    else:
        byc["queries"].update( {"variants": variant_query } )

################################################################################

def create_and_or_query_for_list(logic, q_list):

    if not isinstance(q_list, list):
        return q_list

    if not q_list:
        return {}

    if len(q_list) > 1:
        return { logic: q_list }

    return q_list[0]

################################################################################

def create_in_query_for_parameter(par, qpar, q_pars):

    if not isinstance(q_pars[ par ], list):
        return { qpar: q_pars[ par ] }

    try:
        q_pars[ par ][0]
    except IndexError:
        return { }
 
    if len(q_pars[ par ]) > 1:
        return { qpar: {"$in": q_pars[ par ]} }

    return { qpar: q_pars[ par ][0] }

################################################################################

def variant_create_digest(v, byc):

    t = v["variant_state"]["id"]
    t = re.sub(":", "_", t)

    v_i = v["location"]
    return f'{v_i["chromosome"]}:{v_i["start"]}-{v_i["end"]}:{t}'

################################################################################

def chroname_from_refseqid(refseqid, byc):

    chr_re = re.compile(r'^refseq:NC_0+([^0]\d?)\.\d\d?$')
    chro = chr_re.match(refseqid).group(1)

    return chro

################################################################################

def normalize_pgx_variant(variant, byc, counter=1):
    v_d = byc["variant_definitions"]
    errors = []

    var_id = variant.get("id", counter)

    seq_id = variant["location"].get("sequence_id")
    chromosome = variant["location"].get("chromosome")
    start = variant["location"].get("start")
    end = variant["location"].get("end")
    if not seq_id:
        if chromosome:
            variant["location"].update({"sequence_id": v_d["refseq_aliases"].get(str(chromosome))})
    if not chromosome:
        if seq_id:
            variant["location"].update({"chromosome": v_d["chro_aliases"].get(str(seq_id))})
    if not isinstance(end, int):
        try:
            ref = variant.get("reference_sequence")
            alt = variant.get("sequence")
            v_l = len(ref) - len(alt)
            end_pos = start + abs(v_l) + 1
            # TODO: VRS would do a left-clipping -> start shift ...
            variant["location"].update({"end": end_pos})
        except:
            pass

    # TODO: Some fixes ...
    if "-" in variant.get("sequence", "."):
        variant["sequence"] = re.sub("-", "", variant["sequence"])
    if "-" in variant.get("reference_sequence", "."):
        variant["sequence"] = re.sub("-", "", variant["reference_sequence"])

    var_state_id = variant["variant_state"].get("id")
    variant_type = variant.get("variant_type")
    if not var_state_id:
        if variant_type:
            variant.update({"variant_state": v_d["variant_type_ontologies"][variant_type].get("variant_state")})

    try:
        variant["variant_state"].update({"label": v_d["ontology_variant_types"][var_state_id].get("label")})
    except:
        pass


    for v_l_k in [ "sequence_id", "chromosome", "start", "end" ]:
        if not variant["location"].get(v_l_k):
            errors.append(f'¡¡¡ Parameter `location.{v_l_k}` undefined in variant {var_id} !!!')
    for v_s_k in [ "id", "label" ]:
        if not variant["variant_state"].get(v_s_k):
            errors.append(f'¡¡¡ Parameter `variant_state.{v_s_k}` undefined in variant {var_id} !!!')

    return variant, errors

################################################################################



# __init__.py
import sys
from os import environ, path
from pathlib import Path
import traceback

pkg_path = path.dirname( path.abspath(__file__) )
bycon_lib_path = path.join( pkg_path, "lib" )
sys.path.append( bycon_lib_path )

try:

    from aggregator_utils import *
    from args_parsing import *
    from bycon_helpers import *
    from bycon_plot import *
    from cgi_parsing import *
    from clustering_utils import *
    from data_retrieval import *
    from dataset_parsing import *
    from datatable_utils import *
    from handover_generation import *
    from interval_utils import *
    from file_utils import *
    from filter_parsing import *
    from geomap_utils import *
    from query_execution import *
    from query_generation import *
    from read_specs import *
    from response_remapping import *
    from schema_parsing import *
    from service_utils import *
    from export_file_generation import *
    from variant_parsing import *

    c_f = Path( path.join( pkg_path, "config.yaml" ) )
    config: object = load_yaml_empty_fallback( c_f )
    conf_dir = path.join( pkg_path, "config")
    byc: object = initialize_bycon(config)
    d_f = Path( path.join( conf_dir, "beacon_defaults.yaml" ) )
    byc.update({"beacon_defaults": load_yaml_empty_fallback( d_f ) })
    defaults: object = byc["beacon_defaults"].get("defaults", {})
    for d_k, d_v in defaults.items():
        byc.update( { d_k: d_v } )

    read_bycon_definition_files(conf_dir, byc)
    cgi_parse_query(byc)
    
except Exception:

    if environ.get('HTTP_HOST'):
        print('Content-Type: text/plain')
        print('status: 302')
        print()
    
    print(traceback.format_exc())
    print()
    exit()



#!/usr/bin/env python3

import cgi
import re, json, yaml
from os import environ, pardir, path
import sys, os, datetime

from bycon import *

"""podmd

* <https://progenetix.org/services/dbstats/>
* <https://progenetix.org/services/dbstats/?statsNumber=3&responseFormat=simple>
* <http://progenetix.org/cgi/bycon/services/dbstats.py?method=filtering_terms>

podmd"""

################################################################################
################################################################################
################################################################################

def main():

    try:
        dbstats()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)
    
################################################################################

def dbstats():

    initialize_bycon_service(byc)
    select_dataset_ids(byc)


    if "statsNumber" in byc["form_data"]:
        s_n = byc["form_data"]["stats_number"]
        try:
            s_n = int(s_n)
        except:
            pass
        if type(s_n) == int:
            if s_n > 0:
                byc["stats_number"] = s_n

    create_empty_service_response(byc)
    # exit()

    ds_stats = dbstats_return_latest(byc)

    results = [ ]
    for stat in ds_stats:
        # byc["service_response"]["info"].update({ "date": stat["date"] })
        for ds_id, ds_vs in stat["datasets"].items():
            if len(byc[ "dataset_ids" ]) > 0:
                if not ds_id in byc[ "dataset_ids" ]:
                    continue
            dbs = { "dataset_id": ds_id }
            dbs.update({"counts":ds_vs["counts"]})
            results.append( dbs )

    populate_service_response( byc, results )
    cgi_print_response( byc, 200 )

################################################################################
################################################################################

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# version: 2023-02-13

import sys, re, ruamel.yaml
from os import getlogin, path, system

# from ruamel.yaml.main import round_trip_load, round_trip_dump
# from ruamel.yaml.comments. import CommentedMap

dir_path = path.dirname( path.abspath(__file__) )
pkg_path = path.join( dir_path, "bycon" )
sys.path.append( pkg_path )

from bycon import *

"""
The install script copies the relevant bycon files to the webserver directory
specified in the `config/config.yaml` file and sets the file permissions
accordingly. It requires admin permissions (sudo).

Versions after 2023-02-07 _only_ copy the script directories and package __init__.py
to the server path; additionally a bycon package install is needed with
`pip3 install "bycon>=1.0.16"`.
"""

################################################################################
################################################################################
################################################################################

def main():

    install()

################################################################################
################################################################################
################################################################################

def install():

    yaml = ruamel.yaml.YAML()
    yaml.indent(mapping=2, sequence=4, offset=2)

    i_f = path.join( dir_path, "install.yaml" )
    try:
        with open( i_f ) as y_c:
            install = yaml.load( y_c )
    except Exception as e:
        print(e)
        exit()

    # WARNING: This assumes that the values are sensible...
    for p in ["system_user", "system_group", "bycon_install_dir", "bycon_instance_pars"]:
        p_v = install.get(p, None)
        if p_v is None:
            print("¡¡¡ No `{}` value defined in {} !!!".format(p, i_f))
            exit()

    for p in ["server_tmp_dir_loc", "server_tmp_dir_web"]:
        p_v = install["bycon_instance_pars"].get(p, None)
        if p_v is None:
            print("¡¡¡ No `bycon_instance_pars.{}` value defined in {} !!!".format(p, i_f))
            exit()

    b_i_d_p = path.join( *install["bycon_install_dir"] )
    w_t_d_p = path.join( *install["bycon_instance_pars"]["server_tmp_dir_loc"] )

    for s_p in [b_i_d_p, w_t_d_p]:
        if not path.isdir(s_p):
            print("¡¡¡ {} does not exist - please check & create !!!".format(s_p))
            exit()

    # this block modifies the main config file, so that the `server_tmp_dir_loc`
    # etc. are in line with the install

    c_f = path.join( pkg_path, "config.yaml" )
    c_f_bck = path.join( pkg_path, "config.yaml.bck" )
    try:
        with open( c_f ) as y_c:
            config = yaml.load( y_c )
        system('cp {} {}'.format(c_f, c_f_bck))
    except Exception as e:
        print(e)
        exit()

    for c_p_k, c_p_v in install["bycon_instance_pars"].items():
        config.update({ c_p_k: c_p_v})
    
    with open(c_f, 'w') as out_f:
        yaml.dump(config, out_f)

    # TODO: temporary dev ...
    system(f'sudo rsync -avh --delete {pkg_path}/ {b_i_d_p}/')

    # system('sudo rsync -avh --delete {}/{}/ {}/{}/'.format(pkg_path, "services", b_i_d_p, "services"))
    # system('sudo rsync -avh --delete {}/{}/ {}/{}/'.format(pkg_path, "beaconServer", b_i_d_p, "beaconServer"))
    system('sudo rsync -avh --delete {}/{}/ {}/{}/{}/'.format(dir_path, "local", b_i_d_p, "services", "local"))
    system('sudo rsync -avh --delete {}/{}/ {}/{}/{}/'.format(dir_path, "local", b_i_d_p, "beaconServer", "local"))

    for u_d in install["utility_dirs"]:
        u_p = path.join( *u_d )
        u_p = re.sub("__USERNAME__", getlogin(), u_p)
        if path.isdir(u_p):
            system('rsync -avh --delete {}/{}/ {}/{}/'.format(dir_path, "local", u_p, "local"))
        else:
            print("¡¡¡ Directory {} from install.yaml/#utility_dirs does not exist !!!".format(u_p))
    system('sudo cp {}/{} {}/{}'.format(pkg_path, "__init__.py", b_i_d_p, "__init__.py"))
    system('sudo chown -R {}:{} {}'.format(install["system_user"], install["system_group"], b_i_d_p))
    system('sudo chmod 775 {}/beaconServer/*.py'.format(b_i_d_p))
    system('sudo chmod 775 {}/services/*.py'.format(b_i_d_p))
    system('sudo chmod -R 1777 {}'.format(w_t_d_p))
    print("Updated the `server_tmp_dir_loc` in {} to\n{}".format(c_f, w_t_d_p))
    print("Updated bycon files from\n{}\nto\n{}".format(pkg_path, b_i_d_p))

################################################################################
################################################################################
################################################################################

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import cgi
import re, json
from os import path, pardir
import sys
from pymongo import MongoClient

from bycon import *

"""podmd
* <https://progenetix.org/services/ontologymaps/?filters=NCIT:C3222>
podmd"""

################################################################################
################################################################################
################################################################################

def main():

    try:
        ontologymaps()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)
    
################################################################################

def ontologymaps():

    initialize_bycon_service(byc)

    get_global_filter_flags(byc)
    parse_filters(byc)

    create_empty_service_response(byc)

    q_list = [ ]
    pre_re = re.compile(r'^(\w+?)([:-].*?)?$')
    for f in byc[ "filters" ]:
        f_val = f["id"]
        if pre_re.match( f_val ):
            pre = pre_re.match( f_val ).group(1)

            # TODO TEST
            for f_t, f_d in byc["filter_definitions"].items():
                if re.compile( f_d["pattern"] ).match( f_val ):
                    if "start" in byc[ "filter_flags" ][ "precision" ]:
                        q_list.append( { byc["query_field"]: { "$regex": "^"+f_val } } )
                    elif f["id"] == pre:
                        q_list.append( { byc["query_field"]: { "$regex": "^"+f_val } } )
                    else:
                        q_list.append( { byc["query_field"]: f_val } )

    if len(q_list) < 1:
        response_add_error(byc, 422, "No correct filter value provided!" )

    cgi_break_on_errors(byc)

    query = create_and_or_query_for_list('$and', q_list)

    c_g = [ ]
    u_c_d = { }
    mongo_client = MongoClient( )
    mongo_coll = mongo_client[ byc["config"]["info_db"] ][ byc["ontologymaps_coll"] ]
    for o in mongo_coll.find( query, { '_id': False } ):
        for c in o["code_group"]:
            pre, code = re.split("[:-]", c["id"], maxsplit=1)
            u_c_d.update( { c["id"]: { "id": c["id"], "label": c["label"] } } )
        c_g.append( o["code_group"] )

    u_c = [ ]
    for k, u in u_c_d.items():
        u_c.append(u)
            
    mongo_client.close( )

    results =  [ { "term_groups": c_g, "unique_terms": u_c } ]

# --requestedSchema

    if "termGroups" in byc["response_entity_id"]:
        t_g_s = []
        for tg in c_g:
            t_l = []
            for t in tg:
                t_l.append(str(t.get("id", "")))
                t_l.append(str(t.get("label", "")))
            t_g_s.append("\t".join(t_l))

        if "text" in byc["output"]:
            print_text_response("\n".join(t_g_s))

        results = c_g

    populate_service_response( byc, results)
    # byc["service_response"]["info"]["count"] = len(results[0]["term_groups"])
    cgi_print_response( byc, 200 )

################################################################################
################################################################################

def schema_detyper(parameter):

    if "type" in parameter:
        if "array" in parameter["type"]:
            return [ ]
        elif "object" in parameter["type"]:
            return { }
        return ""

################################################################################
################################################################################

if __name__ == '__main__':
    main()



import re, requests, json, sys, datetime, argparse, urllib.parse, time
from os import path, environ, pardir
from copy import deepcopy
from liftover import get_lifter
from humps import camelize
from cgi_parsing import test_truthy

"""
http://progenetix.test/cgi/bycon/beaconServer/aggregator.py?requestedGranularity=boolean&limit=1000&skip=0&datasetIds=progenetix&assemblyId=GRCh38&referenceName=refseq%3ANC_000009.12&variantType=EFO%3A0030067&filterLogic=AND&includeDescendantTerms=True&start=21000000&start=21975098&end=21967753&end=23000000&filters=NCIT%3AC3058
"""

################################################################################

def set_dataset_id(pvs, ext_defs, ds_id):

    if len(ds_id) < 1:

        return pvs

    if "dataset_ids" in ext_defs["parameter_map"]:
        pvs.update({ ext_defs["parameter_map"]["dataset_ids"].get("remap", "dataset_ids"): ds_id})
        return pvs

    if ext_defs.get("camelize", True) is True: 
        pvs.update({ camelize("dataset_ids"): ds_id})
    else:
        pvs.update({ "dataset_ids": ds_id})

    return pvs

################################################################################

def set_default_values(pvs, ext_defs, byc):

    # adding the parameters with defaults defined in the mappings

    for v_p_k, v_p_v in ext_defs["parameter_map"].items():
        if "default" in v_p_v:
            pvs.update({v_p_v["remap"]: v_p_v["default"]})
    return pvs

################################################################################

def remap_parameters_values(pvs, ext_defs, byc):

    # adding the parameters defined in the mappings
    # TODO: de-convolute; maybe by explicitely having a remap method for each
    #       re-mappable parameter instead of this loop & specials mix

    v_rs_chros = byc["variant_definitions"]["chro_aliases"]
    v_states = byc["variant_definitions"]["variant_state_VCF_aliases"]
    v_p_defs = byc["variant_definitions"]["parameters"]
    service_defaults = byc["service_config"]["beacon_params"].get("defaults", {})

    form_p = deepcopy(byc["form_data"])

    target_assembly = service_defaults.get("assembly_id", "GRCh38")
    if "assembly_id" in ext_defs["parameter_map"]:
        target_assembly = ext_defs["parameter_map"]["assembly_id"].get("value", "GRCh38")

    chro = v_rs_chros.get( form_p.get("reference_name", "NA"), "NA") 

    for v_p_k in v_p_defs.keys():

        if not v_p_k in form_p.keys():
            continue
        val = form_p[v_p_k]
        if v_p_k in ext_defs["parameter_map"].keys():
            v_p_v = ext_defs["parameter_map"][v_p_k]

            if "replace" in v_p_v:
                val = re.sub(v_p_v["replace"][0], v_p_v["replace"][1], val)

            if "reference_name" in v_p_k:
                if "chro" in v_p_v.get("reference_style", ""):
                    val = chro

            if "variant_type" in v_p_k:
                if "VCF" in v_p_v.get("variant_style", "VCF"):
                    if val in v_states:
                        val = v_states[val]

            if "variant_type" in v_p_k:
                if "VCF" in v_p_v.get("variant_style", "VCF"):
                    if val in v_states:
                        val = v_states[val]

            elif "start" in v_p_k or "end" in v_p_k:
                val[0] = int(val[0]) + int(v_p_v.get("shift", 0))
                
                if not "38" in target_assembly:
                    val = lift_positions(target_assembly, chro, val, byc)

            if "array" in v_p_defs[v_p_k].get("type", "string"):
                val = ",".join(map(str, val))

            pvs.update({v_p_v["remap"]: val})

    return pvs

################################################################################

def lift_positions(target_assembly, chro, val, byc):

    l_o_o = byc["service_config"]["liftover_options"]
    s_a = byc["service_config"]["beacon_params"]["defaults"]["assembly_id"]

    # TODO: error, conversion feedback ...
    if target_assembly in l_o_o["supported_sources"].keys():
        return val

    if not target_assembly in l_o_o["supported_targets"].keys():
        return val

    # could be hardcoded; will error out if wrong config
    # a bit of a preparation for swapping input assemblies
    s_l_o_k = l_o_o["supported_sources"][s_a]

    lifted = []

    t_l_o_k = l_o_o["supported_targets"][target_assembly]

    converter = get_lifter(s_l_o_k, t_l_o_k)
    for v in val:
        lifted.append(converter[chro][int(v)][0][1])

    return lifted

################################################################################

def add_parameter_values(pvs, ext_defs, byc):

    # adding the parameters which stay _as is_

    form_p = deepcopy(byc["form_data"])
    v_p_defs = byc["variant_definitions"]["parameters"]
    
    for v_p_k in v_p_defs.keys():
        if not v_p_k in form_p.keys():
            continue
        if v_p_k in ext_defs["parameter_map"].keys():
            continue
        val = form_p[v_p_k]
        if "array" in v_p_defs[v_p_k].get("type", "string"):
            val = ",".join(map(str, val))
        if ext_defs.get("camelize", True) is True: 
            v_p_k = camelize(v_p_k)
        
        pvs.update({v_p_k: val})

    return pvs

################################################################################

def remap_min_max_positions(pvs, ext_defs, byc):

    # special mapping of start | end values for bracket queries to the `startMin`
    # format used in v0.4 / v1 

    form_p = deepcopy(byc["form_data"])

    if ext_defs.get("pos_min_max_remaps", False) is not True:
        return pvs

    if not "start" in form_p or not "end" in form_p:
        return pvs

    if len(form_p["start"]) != 2 or len(form_p["end"]) != 2:
        return pvs

    pvs.update({
        "startMin": form_p["start"][0],
        "startMax": form_p["start"][1],
        "endMin": form_p["end"][0],
        "endMax": form_p["end"][1]
    })

    pvs.pop("start", None)
    pvs.pop("end", None)

    return pvs

################################################################################

def set_fixed_values(pvs, ext_defs, byc):

    for v_p_k, v_p_v in ext_defs["parameter_map"].items():
        if "value" in v_p_v:
            pvs.update({v_p_v["remap"]: v_p_v["value"]})
    return pvs

################################################################################

def retrieve_beacon_response(url, byc):

    try:
        return requests.get(url, headers={ "Content-Type" : "application/json"}, timeout=byc["service_config"]["request_timeout"]).json()
    except requests.exceptions.RequestException as e:
        return { "error": e }

################################################################################

def format_response(r, url, ext_defs, ds_id, byc):

    summary_k = "responseSummary"
    if "response_summary" in ext_defs["response_map"].keys():
        summary_k = ext_defs["response_map"]["response_summary"].get("remap", summary_k)

    if "_root_" in summary_k:
        b_resp = r
    else:
        b_resp = r.get(summary_k, {})

    r = {
            "id": ext_defs["id"],
            "dataset_id": ds_id,
            "api_version": ext_defs.get("api_version", None),
            "exists": test_truthy(b_resp.get("exists", False)),
            "info": deepcopy(ext_defs.get("info", {})),
            "error": str(r.get("error", None)) # str to avoid re-interpretation of code
        }
    r["info"].update({
        "query_url": urllib.parse.unquote(url),
        "query_url_encoded": urllib.parse.quote(url)
    })

    return r

################################################################################
################################################################################
################################################################################

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import cgi
import re, json, yaml
from os import environ, pardir, path, scandir
import sys, datetime
from humps import camelize

from bycon import *

"""podmd

* <https://progenetix.org/services/schemas/biosample>

podmd"""

################################################################################
################################################################################
################################################################################

def main():

    try:
        schemas()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)
    
################################################################################

def schemas():

    initialize_bycon_service(byc)
    create_empty_service_response(byc)

    if "id" in byc["form_data"]:
        schema_name = byc["form_data"].get("id", None)
    else:
        schema_name = rest_path_value("schemas")

    if schema_name is not None:

        comps = schema_name.split('.')
        schema_name = comps.pop(0)

        s = read_schema_file(schema_name, "", byc)
        if s is not False:

            print('Content-Type: application/json')
            print('status:200')
            print()
            print(json.dumps(camelize(s), indent=4, sort_keys=True, default=str)+"\n")
            exit()
    
    response_add_error(byc, 422, "No correct schema id provided!")
    cgi_print_response( byc, 422 )

################################################################################
################################################################################

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import sys, re
from os import path, pardir
from importlib import import_module

from bycon import *
# from bycon.lib.args_parsing import get_bycon_args, args_update_form
# from bycon.lib.cgi_parsing import print_text_response, rest_path_elements, set_debug_state, cgi_print_response
# from bycon.lib.read_specs import load_yaml_empty_fallback, read_bycon_definition_files

"""
"""

################################################################################
################################################################################
################################################################################

def main():

    try:
        beacon()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)

################################################################################

def beacon():

    set_debug_state(debug=0)

    m_f = path.join( pkg_path, "config", "beacon_mappings.yaml")
    b_m = load_yaml_empty_fallback( m_f )

    frm = inspect.stack()[1]
    mod = inspect.getmodule(frm[0])
    if mod is not None:
        sub_path = path.dirname( path.abspath(mod.__file__) )
        conf_dir = path.join( sub_path, "local" )
        if path.isdir(conf_dir):
            m_f = path.join( conf_dir, "beacon_mappings.yaml")
            if path.isfile(m_f):
                b_m = load_yaml_empty_fallback( m_f )
            d_f = Path( path.join( conf_dir, "beacon_defaults.yaml" ) )
            if path.isfile(d_f):
                byc.update({"beacon_defaults": load_yaml_empty_fallback( d_f ) })
                defaults = byc["beacon_defaults"].get("defaults", {})
                for d_k, d_v in defaults.items():
                    byc.update( { d_k: d_v } )
            read_bycon_definition_files(conf_dir, byc)

    byc.update({"request_path_root": "beacon"})
    rest_path_elements(byc)
    r_p_id = byc.get("request_entity_path_id", "info")
    
    get_bycon_args(byc)
    args_update_form(byc)

    if byc["args"]:
        if byc["args"].requestEntityPathId is not None:
            r_p_id = byc["args"].requestEntityPathId

    if r_p_id in b_m["service_aliases"]:
        f = b_m["service_aliases"][ r_p_id ]

        # the data pipeline will run & terminate; else other service
        if f in b_m["data_pipeline_entry_types"]:
            beacon_data_pipeline(byc, f)

        # dynamic package/function loading
        try:
            mod = import_module(f)
            serv = getattr(mod, f)
            serv()
            exit()
        except Exception as e:
            print('Content-Type: text')
            print('status:422')
            print()
            print('Service {} WTF error: {}'.format(f, e))

            exit()

    byc.update({
        "service_response": {},
        "error_response": {
            "error": {
                "error_code": 422,
                "error_message": "No correct service path provided. Please refer to the documentation at http://docs.progenetix.org"
            }
        }
    })

    cgi_print_response(byc, 422)
    
################################################################################
################################################################################

if __name__ == '__main__':
    main()



import pymongo
import re

from bycon.lib.cgi_parsing import rest_path_value


################################################################################

def select_dataset_ids(byc):
    if "dataset_ids" not in byc.keys():
        byc.update({"dataset_ids": []})

    if ds_id_from_rest_path_value(byc) is not False:
        return
    if ds_id_from_accessid(byc) is not False:
        return
    if ds_ids_from_form(byc) is not False:
        return
    if ds_ids_from_args(byc) is not False:
        return

    ################################################################################


def ds_id_from_rest_path_value(byc):
    ds_id = rest_path_value("datasets")

    if ds_id not in byc["dataset_definitions"].keys():
        return False

    byc.update({"dataset_ids": [ds_id]})
    return True


################################################################################

def ds_id_from_accessid(byc):
    # TODO: This is very verbose. In principle there should be an earlier
    # test of existence...

    accessid = byc["form_data"].get("accessid", False)
    info_db = byc["config"].get("info_db", False)
    ho_collname = byc["config"].get("handover_coll", False)

    if any(x is False for x in [accessid, info_db, ho_collname]):
        return False

    ho_client = pymongo.MongoClient()
    h_o = ho_client[info_db][ho_collname].find_one({"id": accessid})
    if not h_o:
        return False

    ds_id = h_o.get("source_db", False)
    if ds_id is False:
        return False

    if ds_id not in byc["dataset_definitions"].keys():
        return False

    byc.update({"dataset_ids": [ds_id]})
    return True


################################################################################

def ds_ids_from_form(byc):
    f_ds_ids = byc["form_data"].get("dataset_ids", False)
    if f_ds_ids is False:
        return False

    ds_ids = []
    for ds_id in f_ds_ids:
        if ds_id in byc["dataset_definitions"].keys():
            ds_ids.append(ds_id)

    if len(ds_ids) < 1:
        return False

    byc.update({"dataset_ids": ds_ids})
    return True


################################################################################

def ds_ids_from_args(byc):
    if "args" not in byc or byc["args"] is None:
        return False

    if byc["args"].datasetIds:
        ds_ids = re.split(",", byc["args"].datasetIds)
        byc.update({"dataset_ids": ds_ids})
        return True

    return False



import csv
import datetime
import re
import requests
from pathlib import Path
from os import path
from copy import deepcopy
from random import sample as random_samples

from cgi_parsing import prjsonnice
from datatable_utils import import_datatable_dict_line
from interval_utils import interval_cnv_arrays, interval_counts_from_callsets
from response_remapping import de_vrsify_variant, vrsify_variant
from variant_parsing import variant_create_digest


################################################################################

def read_tsv_to_dictlist(filepath, max_count=0):

    dictlist = []

    with open(filepath, newline='') as csvfile:
    
        data = csv.DictReader(filter(lambda row: row[0]!='#', csvfile), delimiter="\t", quotechar='"')
        fieldnames = list(data.fieldnames)

        for l in data:
            dictlist.append(dict(l))

    if 0 < max_count < len(dictlist):
        dictlist = random_samples(dictlist, k=max_count)

    return dictlist, fieldnames


################################################################################

def read_www_tsv_to_dictlist(www, max_count=0):

    dictlist = []

    with requests.Session() as s:
        download = s.get(www)
        decoded_content = download.content.decode('utf-8')    
        data = csv.DictReader(filter(lambda row: row[0]!='#', decoded_content.splitlines()), delimiter="\t", quotechar='"')
        fieldnames = list(data.fieldnames)

        for l in data:
            dictlist.append(dict(l))

    if 0 < max_count < len(dictlist):
        dictlist = random_samples(dictlist, k=max_count)

    return dictlist, fieldnames

################################################################################

def callset_guess_probefile_path(callset, byc):

    d = Path( path.join( *byc["config"]["server_callsets_dir_loc"]))
    n = byc["config"].get("callset_probefile_name", "___none___")

    if not d.is_dir():
        return False

    if not "analysis_info" in callset:
        return False

    # TODO: not only geo
    s_id = callset["analysis_info"].get("series_id", "___none___").replace("geo:", "")
    e_id = callset["analysis_info"].get("experiment_id", "___none___").replace("geo:", "")

    p_f = Path( path.join( d, s_id, e_id, n ) )

    if not p_f.is_file():
        return False

    return p_f


################################################################################
################################################################################
################################################################################

class ByconBundler:

    """
    # The `ByconBundler` class

    This class bundles documents from the main entities which have a complete
    intersection - e.g. for a set of variants their callsets, biosamples and
    individuals. The bundling does _not_ have to be complete; e.g. bundles may
    be based on only some matched variants (not all variants of the referenced
    callsets); and bundles may have empty lists for some entities.
    """

    def __init__(self, byc):

        self.byc = byc
        self.errors = []
        self.filepath = None
        self.header = []
        self.data = []
        self.fieldnames = []
        self.callsetVariantsBundles = []
        self.intervalFrequenciesBundles = []

        self.bundle = {
            "variants": [],
            "callsets": [],
            "biosamples": [],
            "individuals": [],
            "ds_id": None,
            "info": {
                "errors": []
            }
        }

        self.keyedBundle = {
            "variants_by_callset_id": {},
            "callsets_by_id": {},
            "individuals_by_id": {},
            "biosamples_by_id": {},
            "ds_id": None,
            "info": {
                "errors": []
            }
        }

    #--------------------------------------------------------------------------#
    #----------------------------- public -------------------------------------#
    #--------------------------------------------------------------------------#

    def read_pgx_file(self, filepath):

        self.filepath = filepath

        h_lines = []

        with open(self.filepath) as f:
            for line in f:
                line = line.strip()
                if line.startswith("#"):
                    h_lines.append(line)

        d_lines, fieldnames = read_tsv_to_dictlist(self.filepath, max_count=0)

        self.header = h_lines
        self.data = d_lines
        self.fieldnames = fieldnames

        return self
        

    #--------------------------------------------------------------------------#

    def read_probedata_file(self, filepath):

        self.filepath = filepath
        self.probedata = []

        p_lines, fieldnames = read_tsv_to_dictlist(self.filepath, max_count=0)

        p_o = {
            "probe_id": False,
            "reference_name": False,
            "start": False,
            "value": False
        }

        p_f_d = {
            "probe_id": {"type": "string", "key": fieldnames[0]},
            "reference_name": {"type": "string", "key": fieldnames[1]},
            "start": {"type": "integer", "key": fieldnames[2]},
            "value": {"type": "number", "key": fieldnames[3]}
        }

        for l in p_lines:
            p = deepcopy(p_o)
            for pk, pv in p_f_d.items():
                l_k = pv["key"]
                p.update({ pk: l.get(l_k) })
                if "int" in pv["type"]:
                    p.update({ pk: int(p[pk]) })
                elif "num" in pv["type"]:
                    p.update({ pk: float(p[pk]) })
            self.probedata.append(p)

        return self.probedata

    #--------------------------------------------------------------------------#

    def pgxseg_to_keyed_bundle(self, filepath):

        self.read_pgx_file(filepath)


        if not "biosample_id" in self.fieldnames:
            self.errors.append("¡¡¡ The `biosample_id` parameter is required for variant assignment !!!")
            return

        self.__deparse_pgxseg_samples_header()
        self.__keyed_bundle_add_variants()

        return self.keyedBundle

    #--------------------------------------------------------------------------#

    def pgxseg_to_bundle(self, filepath):

        self.pgxseg_to_keyed_bundle(filepath)
        self.__flatten_keyed_bundle()

        return self.bundle

    #--------------------------------------------------------------------------#

    def callsets_variants_bundles(self):

        # TODO: This is similar to a keyed bundle component ...

        bb = self.bundle

        c_p_l = []
        for p_o in bb.get("callsets", []):
            cs_id = p_o.get("id")
            p_o.update({
                "ds_id": bb.get("ds_id", ""),
                "variants":[]
            })

            for v in bb["variants"]:
                if v.get("callset_id", "") == cs_id:
                    v = de_vrsify_variant(v, self.byc)
                    p_o["variants"].append(v)

            c_p_l.append(p_o)
            
        self.callsetVariantsBundles = c_p_l

        return self.callsetVariantsBundles

    #--------------------------------------------------------------------------#

    def callsets_frequencies_bundles(self):
            
        self.intervalFrequenciesBundles.append(self.__callsetBundleCreateIset("import"))

        return self.intervalFrequenciesBundles

    #--------------------------------------------------------------------------#
    #----------------------------- private ------------------------------------#
    #--------------------------------------------------------------------------#

    def __deparse_pgxseg_samples_header(self):

        b_k_b = self.keyedBundle
        h_l = self.header

        for l in h_l:
            if not l.startswith("#sample=>"):
                continue       
            l = re.sub("#sample=>", "", l)
            bios_d = {}
            for p_v in l.split(";"):
                k, v = p_v.split("=")
                v = re.sub(r'^[\'\"]', '', v)
                v = re.sub(r'[\'\"]$', '', v)
                bios_d.update({k:v})
            fieldnames = list(bios_d.keys())
            bs_id = bios_d.get("biosample_id")
            if bs_id is None:
                continue

            bios = {"id": bs_id} 
            bios = import_datatable_dict_line(self.byc, bios, fieldnames, bios_d, "biosample")
            cs_id = bios.get("callset_id", re.sub("pgxbs", "pgxcs", bs_id) )
            ind_id = bios.get("individual_id", re.sub("pgxbs", "pgxind", bs_id) )
            ind = {"id": ind_id} 
            cs = {"id": cs_id, "biosample_id": bs_id, "individual_id": ind_id} 

            bios.update({"individual_id": ind_id})

            b_k_b["callsets_by_id"].update({ cs_id: import_datatable_dict_line(self.byc, cs, fieldnames, bios_d, "analysis") })
            b_k_b["individuals_by_id"].update({ ind_id: import_datatable_dict_line(self.byc, ind, fieldnames, bios_d, "individual") })
            b_k_b["biosamples_by_id"].update({ bs_id: bios })
            b_k_b["variants_by_callset_id"].update({ cs_id: [] })

        self.keyedBundle = b_k_b

    #--------------------------------------------------------------------------#

    def __keyed_bundle_add_variants(self):

        fieldnames = self.fieldnames
        varlines = self.data

        b_k_b = self.keyedBundle

        inds_ided = b_k_b.get("individuals_by_id", {})
        bios_ided = b_k_b.get("biosamples_by_id", {})
        cs_ided = b_k_b.get("callsets_by_id", {})
        vars_ided = b_k_b.get("variants_by_callset_id", {})

        for c, v in enumerate(varlines):
            bs_id = v.get("biosample_id", False)

            # If the biosample exists in metadata all the other items will exist by id
            if not bs_id in bios_ided:
                cs_id = re.sub(r'^(pgxbs-)?', "pgxcs-", bs_id)
                ind_id = re.sub(r'^(pgxbs-)?', "pgxind-", bs_id)
                cs_ided.update( {cs_id: {"id": cs_id, "biosample_id": bs_id, "individual_id": ind_id } } )
                vars_ided.update( {cs_id: [] } )
                bios_ided.update( {bs_id: {"id": bs_id, "individual_id": ind_id } } )
                inds_ided.update( {ind_id: {"id": ind_id } } )
            else:
                for cs_i, cs_v in cs_ided.items():
                    if cs_v.get("biosample_id", "___nothing___") == bs_id:
                        cs_id = cs_i
                        continue
            
            bios = bios_ided.get(bs_id)
            cs = cs_ided.get(cs_id)
            ind_id = bios.get("individual_id", "___nothing___")
            ind = inds_ided.get(ind_id)

            update_v = {
                "individual_id": ind_id,
                "biosample_id": bs_id,
                "callset_id": cs_id,
            }

            update_v = import_datatable_dict_line(self.byc, update_v, fieldnames, v, "variant")
            vrsify_variant(update_v, self.byc)
            update_v.update({
                "variant_internal_id": variant_create_digest(update_v, self.byc),
                "updated": datetime.datetime.now().isoformat()
            })

            vars_ided[cs_id].append(update_v)

        for cs_id, cs_vars in vars_ided.items():
            maps, cs_cnv_stats, cs_chro_stats = interval_cnv_arrays(cs_vars, self.byc)           
            cs_ided[cs_id].update({"cnv_statusmaps": maps})
            cs_ided[cs_id].update({"cnv_stats": cs_cnv_stats})
            cs_ided[cs_id].update({"cnv_chro_stats": cs_chro_stats})
            cs_ided[cs_id].update({"updated": datetime.datetime.now().isoformat()})

        self.keyedBundle.update({
            "individuals_by_id": inds_ided,
            "biosamples_by_id": bios_ided,
            "callsets_by_id": cs_ided,
            "variants_by_callset_id": vars_ided
        })

    #--------------------------------------------------------------------------#

    def __flatten_keyed_bundle(self):

        b_k_b = self.keyedBundle

        bios_k = b_k_b.get("biosamples_by_id", {})
        ind_k = b_k_b.get("individuals_by_id", {})
        cs_k = b_k_b.get("callsets_by_id", {})
        v_cs_k = b_k_b.get("variants_by_callset_id", {})

        self.bundle.update({
            "biosamples": list( bios_k.values() ),
            "individuals": list( ind_k.values() ),
            "callsets": list( cs_k.values() ),
            "variants": [elem for sublist in ( v_cs_k.values() ) for elem in sublist]
        })

    #--------------------------------------------------------------------------#

    def __callsetBundleCreateIset(self, label=""):

        intervals, cnv_cs_count = interval_counts_from_callsets(self.bundle["callsets"], self.byc)

        ds_id = self.bundle.get("ds_id", "")
        iset = {
            "dataset_id": ds_id,
            "group_id": ds_id,
            "label": label,
            "sample_count": cnv_cs_count,
            "interval_frequencies": []
        }

        for intv_i, intv in enumerate(intervals):
            iset["interval_frequencies"].append(intv.copy())

        return iset

################################################################################



#!/usr/bin/env python3

from os import path, pardir, environ
import sys, re
from importlib import import_module

from bycon import *
pkg_path = path.join( path.dirname( path.abspath(__file__) ), pardir, pardir )

"""
The `services` application deparses a request URI and calls the respective
script. The functionality is combined with the correct configuration of a 
rewrite in the server configuration for creation of canonical URLs.
"""

################################################################################
################################################################################
################################################################################

def main():

    try:
        services()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)
    
################################################################################

def services():

    set_debug_state(debug=0)

    m_f = path.join( pkg_path, "config", "services_mappings.yaml")
    b_m = load_yaml_empty_fallback( m_f )

    frm = inspect.stack()[1]
    mod = inspect.getmodule(frm[0])
    if mod is not None:
        sub_path = path.dirname( path.abspath(mod.__file__) )
        conf_dir = path.join( sub_path, "local" )
        if path.isdir(conf_dir):
            m_f = path.join( conf_dir, "services_mappings.yaml")
            if path.isfile(m_f):
                b_m = load_yaml_empty_fallback( m_f )
            d_f = Path( path.join( conf_dir, "beacon_defaults.yaml" ) )
            if path.isfile(d_f):
                byc.update({"beacon_defaults": load_yaml_empty_fallback( d_f ) })
                defaults = byc["beacon_defaults"].get("defaults", {})
                for d_k, d_v in defaults.items():
                    byc.update( { d_k: d_v } )
            read_bycon_definition_files(conf_dir, byc)

    byc.update({"request_path_root": "services"})
    rest_path_elements(byc)
    r_p_id = byc.get("request_entity_path_id", "__empty_value__")

    if r_p_id in b_m["service_aliases"]:    
        f = b_m["service_aliases"][ r_p_id ]
        
        # dynamic package/function loading
        try:
            mod = import_module(f)
            serv = getattr(mod, f)
            serv()
            exit()
        except Exception as e:
            print('Content-Type: text')
            print('status:422')
            print()
            print('Service {} error: {}'.format(f, e))

            exit()

    if r_p_id in b_m["rewrites"]:
        uri = environ.get('REQUEST_URI')
        pat = re.compile( rf"^.+\/{r_p_id}\/?(.*?)$" )
        if pat.match(uri):
            stuff = pat.match(uri).group(1)
            print_uri_rewrite_response(b_m["rewrites"][r_p_id], stuff)

    byc.update({
        "service_response": {},
        "error_response": {
            "error": {
                "error_code": 422,
                "error_message": "No correct service path provided. Please refer to the documentation at http://docs.progenetix.org"
            }
        }
    })

    cgi_print_response(byc, 422)

################################################################################
################################################################################

if __name__ == '__main__':
    main()



import csv, re, requests
# from attrdictionary import AttrDict
from random import sample as randomSamples

from cgi_parsing import prjsonnice

################################################################################

def export_datatable_download(results, byc):

    # TODO: separate table generation from HTTP response

    if not "table" in byc["output"]:
        return
    if not "datatable_mappings" in byc:
        return

    dt_m = byc["datatable_mappings"]
    r_t = byc["response_entity_id"]
    io_params = dt_m["entities"][ r_t ]["parameters"]

    if not byc["response_entity_id"] in dt_m["entities"]:
        return

    if not "local" in byc["env"]:
 
        print('Content-Type: text/tsv')
        if byc["download_mode"] is True:
            print('Content-Disposition: attachment; filename='+byc["response_entity_id"]+'.tsv')
        print('status: 200')
        print()

    if "idtable" in byc["output"]:
        io_params = {"id": {"db_key":"id", "type": "string" } }

    header = create_table_header(io_params)

    print("\t".join( header ))

    for pgxdoc in results:

        line = [ ]

        for par, par_defs in io_params.items():

            parameter_type = par_defs.get("type", "string")
            pres = par_defs.get("prefix_split", {})

            if len(pres.keys()) < 1:
                db_key = par_defs.get("db_key", "___undefined___")
                v = get_nested_value(pgxdoc, db_key)
                if isinstance(v, list):
                    line.append("::".join(map(str, (v))))
                else:
                    line.append(str(v))
            else:

                par_vals = pgxdoc.get(par, [])

                # TODO: this is based on the same order of prefixes as in the
                # header generation, w/o checking; should be keyed ...
                for pre in pres.keys():
                    p_id = ""
                    p_label = ""
                    for v in par_vals:
                        if v.get("id", "___none___").startswith(pre):
                            p_id = v.get("id")
                            p_label = v.get("label", "")
                            continue
                    line.append(p_id)
                    line.append(p_label)

        print("\t".join( line ))

    exit()

################################################################################

def import_datatable_dict_line(byc, parent, fieldnames, lineobj, primary_scope="biosample"):

    dt_m = byc["datatable_mappings"]

    io_params = dt_m["entities"][ primary_scope ]["parameters"]
    def_params = create_table_header(io_params)

    pref_array_values = {}

    for f_n in fieldnames:

        if "#"in f_n:
            continue

        if f_n not in def_params:
            continue

        # this is for the split-by-prefix columns
        par = re.sub(r'___.*?$', '', f_n)
        par_defs = io_params.get(par, {})

        v = lineobj[f_n].strip()

        if len(v) < 1:
            if f_n in io_params.keys():
                v = io_params[f_n].get("default", "")

        if len(v) < 1:
            continue

        # this makes only sense for updating existing data; if there would be
        # no value, the parameter would just be excluded from the update object
        # if there was an empy value
        if "__delete__" in v.lower():
            v = ""

        parameter_type = par_defs.get("type", "string")
        if "num" in parameter_type:
            v = float(v)
        elif "integer" in parameter_type:
            v = int(v)

        if re.match(r"^(\w+[a-zA-Z0-9])_(id|label)___(\w+)$", f_n):
            p, key, pre = re.match(r"^(\w+)_(id|label)___(\w+)$", f_n).group(1,2,3)
            # TODO: this is a bit complicated - label and id per prefix ...
            if not p in pref_array_values.keys():
                pref_array_values.update({p:{pre:{}}})
            if not pre in pref_array_values[p].keys():
                pref_array_values[p].update({pre:{}})
            pref_array_values[p][pre].update({key:v})
            continue
        
        if par in io_params.keys():
            db_key = io_params[par]["db_key"]

            # assign_nested_attribute(parent, db_key, v)
            assign_nested_value(parent, db_key, v, parameter_type)

    for l_k, pre_item in pref_array_values.items():
        if not l_k in parent:
            parent.update({l_k:[]})
        for pre, p_v in pre_item.items():
            parent[l_k].append(p_v)

    return parent

################################################################################

def create_table_header(io_params):

    """podmd
    podmd"""

    header_labs = [ ]

    for par, par_defs in io_params.items():

        pres = par_defs.get("prefix_split", {})

        if len(pres.keys()) < 1:
            header_labs.append( par )
            continue

        for pre in pres.keys():
            header_labs.append( par+"_id"+"___"+pre )
            header_labs.append( par+"_label"+"___"+pre )

    return header_labs

################################################################################

# def assign_nested_attribute(parent, dotted_key, v):

#     n_p = AttrDict(parent)
#     keys = dotted_key.split('.')
#     for key in keys[:-1]:
#         n_p = getattr(n_p, key)
#     setattr(n_p, keys[-1], v)

#     return dict(parent)

# ################################################################################

# def get_nested_value(obj, dot_string):
#     keys = dot_string.split('.')
#     for key in keys:
#         obj = getattr(obj, key)
#     return obj

################################################################################

def assign_nested_value(parent, dotted_key, v, parameter_type="string"):

    if v is None:
        return parent

    if "num" in parameter_type:
        v = float(v)
    elif "integer" in parameter_type:
        v = int(v)
    else:
        v = str(v)

    ps = dotted_key.split('.')

    if len(ps) == 1:
        if "array" in parameter_type:
            parent.update({ps[0]: v.split(',')})
        else:
            parent.update({ps[0]: v })
        return parent

    if ps[0] not in parent or parent[ ps[0] ] is None:
        parent.update({ps[0]: {}})

    if len(ps) == 2:
        if "array" in parameter_type:
            parent[ ps[0] ].update({ps[1]: v.split(',')})
        else:
            parent[ ps[0] ].update({ps[1]: v })
        return parent

    if  ps[1] not in parent[ ps[0] ] or parent[ ps[0] ][ ps[1] ] is None:
        parent[ ps[0] ].update({ps[1]: {}})
    if len(ps) == 3:
        if "array" in parameter_type:
            parent[ ps[0] ][ ps[1] ].update({ps[2]: v.split(',')})
        else:
            parent[ ps[0] ][ ps[1] ].update({ps[2]: v })
        return parent

    if  ps[2] not in parent[ ps[0] ][ ps[1] ] or parent[ ps[0] ][ ps[1] ][ ps[2] ] is None:
        parent[ ps[0] ][ ps[1] ].update({ps[2]: {}})
    if len(ps) == 4:
        if "array" in parameter_type:
            parent[ ps[0] ][ ps[1] ][ ps[2] ].update({ps[3]: v.split(',')})
        else:
            parent[ ps[0] ][ ps[1] ][ ps[2] ].update({ps[3]: v })
        return parent
    
    if len(ps) > 4:
        print("¡¡¡ Parameter key "+dotted_key+" nested too deeply (>4) !!!")
        return '_too_deep_'

    return parent

################################################################################

def get_nested_value(parent, dotted_key):

    ps = dotted_key.split('.')

    v = ""

    if len(ps) == 1:
        try:
            v = parent[ ps[0] ]
        except:
            v = ""
    elif len(ps) == 2:
        try:
            v = parent[ ps[0] ][ ps[1] ]
        except:
            v = ""
    elif len(ps) == 3:
        try:
            v = parent[ ps[0] ][ ps[1] ][ ps[2] ]
        except:
            v = ""
    elif len(ps) == 4:
        try:
            v = parent[ ps[0] ][ ps[1] ][ ps[2] ][ ps[3] ]
        except:
            v = ""
    elif len(ps) == 5:
        try:
            v = parent[ ps[0] ][ ps[1] ][ ps[2] ][ ps[3] ][ ps[4] ]
        except:
            v = ""
    elif len(ps) > 5:
        print("¡¡¡ Parameter key "+dotted_key+" nested too deeply (>5) !!!")
        return '_too_deep_'

    return v

################################################################################



#!/usr/bin/env python3

import cgi
import re, json, yaml
from os import path, environ, pardir
import sys, datetime, argparse
from pathlib import Path

from bycon import *

"""podmd

podmd"""

################################################################################
################################################################################
################################################################################

def main():

    try:
        samples_plotter()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)

################################################################################

def samplesPlotter():
    
    try:
        samples_plotter()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)
   
################################################################################

def samples_plotter():

    initialize_bycon_service(byc)
    parse_variant_parameters(byc)
    generate_genomic_intervals(byc)

    byc["response_entity"].update({"response_schema": "beaconResultsetsResponse"})
    create_empty_beacon_response(byc)

    id_rest = rest_path_value("samplesPlotter")

    if id_rest is not None:
        byc[ "file_id" ] = id_rest
    elif "file_id" in byc["form_data"]:
        byc[ "file_id" ] = byc["form_data"]["file_id"]

    if not "file_id" in byc:
        response_add_error(byc, 422, "No value was provided for collation `fileId`.")
    cgi_break_on_errors(byc)

    _create_file_handover_response(byc)

    # TODO: maybe move to defaults file...
    if not "plot" in byc.get("output", "histoplot"):
        byc.update({"output": "histoplot"})

    inputfile = path.join( *config[ "server_tmp_dir_loc" ], byc[ "file_id" ] )

    if not path.exists(inputfile):
        response_add_error(byc, 422, f"The file path {inputfile} does not exist.")  
    cgi_break_on_errors(byc)

    pb = ByconBundler(byc)
    pb.pgxseg_to_bundle(inputfile)
    plot_data_bundle = {
        "interval_frequencies_bundles": pb.callsets_frequencies_bundles(),
        "callsets_variants_bundles": pb.callsets_variants_bundles()
    }

    ByconPlot(byc, plot_data_bundle).svg_response()

################################################################################

def _create_file_handover_response(byc):

    if "plot" in byc.get("output", "handovers"):
        return

    h_o_types = byc["handover_definitions"]["h->o_types"]
    h_o_server = select_this_server(byc)
    res_h_o = []

    for h_o_t in ["histoplot", "samplesplot"]:
        h_o_defs = h_o_types.get(h_o_t, {})
        h_o_defs.update({"script_path_web":"/services/samplesPlotter"})
        h_o_r = {
            "handover_type": {
                "id": h_o_defs[ "id" ],
                "label": f'{ h_o_defs.get("label", "generic handover") }'
            },
            "description": h_o_defs[ "description" ],
            "url": handover_create_url(h_o_server, h_o_defs, byc[ "file_id" ], byc),
            "pages": []
        }
        h_o_r.update({"url": re.sub("accessid", "fileId", h_o_r["url"])})
        res_h_o.append(h_o_r)

    r_set = { "exists": True, "resultsHandovers": res_h_o }

    byc["service_response"].update({
        "response_summary":{"exists":True},
        "response": {"result_sets": [r_set]}
    })

    cgi_print_response( byc, 200 )

################################################################################
################################################################################
################################################################################

if __name__ == '__main__':
    main()



import base36
import time
import re
from isodate import parse_duration

################################################################################

def generate_id(prefix):
    time.sleep(.001)
    return '{}-{}'.format(prefix, base36.dumps(int(time.time() * 1000)))  ## for time in ms


################################################################################

def days_from_iso8601duration(iso8601duration):
    """A simple function to convert ISO8601 duration strings to days. This is
    potentially lossy since it does not include time parsing."""

    # TODO: check format
    is_isodate_duration = re.match(r'^P\d+?[YMD](\d+?[M])?(\d+?[D])?', iso8601duration)
    if not is_isodate_duration:
        return False

    duration = parse_duration(iso8601duration)
    days = 0
    try:
        days += int(duration.years) * 365.2425
    except AttributeError:
        pass
    try:
        days += int(duration.months) * 30.4167
    except AttributeError:
        pass
    try:
        days += int(duration.days)
    except AttributeError:
        pass

    return days


################################################################################

def hex_2_rgb( hexcolor ):

    rgb = [127, 127, 127]
    h = hexcolor.lstrip('#')
    rgb = tuple(int(h[i:i+2], 16) for i in (0, 2, 4))

    return rgb





"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="bycon",
    version="1.0.54",
    description="A Python-based environment for the Beacon v2 genomics API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/progenetix/bycon",
    author="Michael Baudis",
    author_email="contact@progenetix.org",
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only"
    ],
    keywords="genomics, Beacon",
    package_dir={'':"."},
    packages=find_packages(where="."),
    setup_requires=['setuptools_scm'],
    include_package_data=True,
    package_data={'bycon': ['**/*.yaml', '**/*.txt', '**/*.json']},
    python_requires=">=3.7, <4",
    install_requires=["base36", "isodate", "Cython", "json_ref_dict", "liftover", "numpy", "progress", "pyhumps", "pymongo", "PyYAML", "ruamel.yaml"],  # Optional
    project_urls={  # Optional
        "Bug Reports": "https://github.com/progenetix/bycon/issues",
        "Source": "https://github.com/progenetix/bycon/",
    },
)


import datetime, re
from pymongo import MongoClient
from cgi_parsing import *
from interval_utils import interval_counts_from_callsets


################################################################################

def reshape_resultset_results(ds_id, r_s_res, byc):
    r_s_res = remap_variants_to_VCF(r_s_res, byc)
    r_s_res = remap_variants(r_s_res, byc)
    r_s_res = remap_analyses(r_s_res, byc)
    r_s_res = remap_biosamples(r_s_res, byc)
    r_s_res = remap_cohorts(r_s_res, byc)
    r_s_res = remap_individuals(r_s_res, byc)
    r_s_res = remap_phenopackets(ds_id, r_s_res, byc)
    r_s_res = remap_runs(r_s_res, byc)
    r_s_res = remap_all(r_s_res)

    return r_s_res


################################################################################

def remap_variants(r_s_res, byc):
    """
    Since the Beacon default model works from the concept of a canonical genomic
    variation and the bycon data model uses variant instances, the different
    instances of a "canonical" variant have to be identified and grouped together
    with individual instances indicated through their identifiers in `caseLevelData`.
    """

    if not "variant" in byc["response_entity_id"].lower():
        return r_s_res
    if "vcf" in byc["output"].lower():
        return r_s_res

    v_d = byc["variant_definitions"]

    variant_ids = []
    for v in r_s_res:
        variant_ids.append(v["variant_internal_id"])
    variant_ids = list(set(variant_ids))

    variants = []

    for d in variant_ids:

        d_vs = [var for var in r_s_res if var.get('variant_internal_id', "__none__") == d]

        v = {
            "variant_internal_id": d,
            "variation": d_vs[0], "case_level_data": []
        }

        v["variation"].pop("variant_internal_id", None)

        for d_v in d_vs:
            v["case_level_data"].append(
                {
                    "id": d_v.get("id", "__none__"),
                    "biosample_id": d_v.get("biosample_id", "__none__"),
                    "analysis_id": d_v.get("callset_id", "__none__")
                }
            )

        # TODO: Keep legacy pars?
        legacy_pars = ["_id", "id", "reference_name", "type", "biosample_id", "callset_id", "individual_id",
                       "variant_type", "reference_bases", "alternate_bases", "start", "end"]
        for p in legacy_pars:
            v["variation"].pop(p, None)

        variants.append(v)

    return variants


################################################################################

def remap_variants_to_VCF(r_s_res, byc):
    """
    """

    if not "variant" in byc["response_entity_id"].lower():
        return r_s_res
    if not "vcf" in byc["output"].lower():
        return r_s_res

    open_text_streaming(byc["env"], "variants.vcf")
    print(
        """##fileformat=VCFv4.4
##reference=GRCh38
##ALT=<ID=DUP,Description="Duplication">
##ALT=<ID=DEL,Description="Deletion">
##INFO=<ID=END,Number=1,Type=Integer,Description="End position of the longest variant described in this record">
##INFO=<ID=SVLEN,Number=A,Type=Integer,Description="Length of structural variant">
##INFO=<ID=CN,Number=A,Type=Float,Description="Copy number of CNV/breakpoint">
##INFO=<ID=SVCLAIM,Number=A,Type=String,Description="Claim made by the structural variant call. Valid values are D, J, DJ for abundance, adjacency and both respectively">
##INFO=<ID=IMPRECISE,Number=0,Type=Flag,Description="Imprecise structural variation">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">"""
    )

    v_d = byc["variant_definitions"]
    v_o = {
        "#CHROM": ".",
        "POS": ".",
        "ID": ".",
        "REF": ".",
        "ALT": ".",
        "QUAL": ".",
        "FILTER": "PASS",
        "FORMAT": "",
        "INFO": ""
    }

    variant_ids = ()

    sorted_vars = list(sorted(r_s_res, key=lambda x: (f'{x["location"]["chromosome"].replace("X", "XX").replace("Y", "YY").zfill(2)}', x["location"]['start'])))

    for v in sorted_vars:
        variant_ids += (v.get("variant_internal_id", "__none__"), )

    biosample_ids = []
    for v in sorted_vars:
        biosample_ids.append(v.get("biosample_id", "__none__"))

    biosample_ids = list(set(biosample_ids))

    for bsid in biosample_ids:
        v_o.update({bsid: "."})

    variants = []
    print("   ".join(v_o.keys()))

    for d in variant_ids:

        d_vs = [var for var in sorted_vars if var.get('variant_internal_id', "__none__") == d]

        vcf_v = vcf_variant(d_vs[0], v_o.copy(), byc)

        for bsid in biosample_ids:
            vcf_v.update({bsid: "."})

        for d_v in d_vs:
            b_i = d_v.get("biosample_id", "__none__")
            vcf_v.update({b_i: "0/1"})

        r_l = map(str, list(vcf_v.values()))
        # variants.append("   ".join(r_l))
        print("   ".join(r_l))

    exit()
    return variants


################################################################################

def vrsify_variant(variant, byc):
    # TODO: Still prototyping; also need to accomodate VRS 1.3

    v_d = byc["variant_definitions"]
    v_t_als = v_d["variant_state_aliases"]
    r_a = v_d["refseq_aliases"]
    efo_vrs = v_d["ontology_variant_types"]

    v_type = variant.get("variant_type", False)
    if v_type is False:
        return variant

    if v_type not in v_t_als:
        return variant

    variant.update({
        "variant_state": {
            "id": v_t_als[v_type],
            "label": efo_vrs[v_t_als[v_type]].get("label", "")
        }})

    sequence_id = r_a.get(variant.get("reference_name", "___none___"), False)
    if sequence_id is False:
        return variant

    variant["location"].update({"sequence_id": sequence_id})

    return variant


################################################################################

def vcf_variant(v, vcf_v, byc):
    v = de_vrsify_variant(v, byc)
    if v is False:
        return v

    vcf_v.update({
        "#CHROM": v["reference_name"],
        "POS": v["start"],
        "ID": ".",
        "REF": v["reference_bases"],
        "ALT": v["alternate_bases"],
        "QUAL": ".",
        "FILTER": "PASS",
        "FORMAT": "GT",
        "INFO": ""
    })

    v_l = v["variant_label"]
    v_s =  v["variant_type"]

    # TODO: not very general...
    if "copy" in v_l:
        vcf_v.update({"ALT": f'<{v["variant_type"]}>'})
        v_l =  v["end"] - v["start"]
        vcf_v.update({"INFO": f'IMPRECISE;SVCLAIM=D;END={v["end"]};SVLEN={v_l}'})

    return vcf_v


################################################################################

def de_vrsify_variant(v, byc):
    v_d = byc["variant_definitions"]

    r_n = v["location"].get("sequence_id")
    if r_n is None:
        return False

    v_r = {
        "id": v.get("id"),
        "variant_internal_id": v.get("variant_internal_id"),
        "callset_id": v.get("callset_id"),
        "biosample_id": v.get("biosample_id"),
        "reference_bases": v.get("reference_sequence", "."),
        "alternate_bases": v.get("sequence", "."),
        "reference_name": v["location"].get("chromosome", "."),
        "sequence_id": r_n,
        "start": v["location"]["start"],
        "end": v["location"]["end"],
        "info": v.get("info", {}),
        "variant_label": v["variant_state"].get("label", "")
    }

    efo = v["variant_state"].get("id")
    v_r.update({"variant_type": v_d["ontology_variant_types"][efo]["DUPDEL"]})

    return v_r


################################################################################

def remap_analyses(r_s_res, byc):
    if not "analysis" in byc["response_entity_id"]:
        return r_s_res

    pop_keys = ["info", "provenance", "cnv_statusmaps", "cnv_chro_stats", "cnv_stats"]

    if "stats" in byc["output"]:
        pop_keys = ["info", "provenance", "cnv_statusmaps"]

    for cs_i, cs_r in enumerate(r_s_res):
        # TODO: REMOVE VERIFIER HACKS
        r_s_res[cs_i].update({"pipeline_name": "progenetix", "analysis_date": "1967-11-11"})
        for k in pop_keys:
            r_s_res[cs_i].pop(k, None)

    return r_s_res


################################################################################

def remap_cohorts(r_s_res, byc):
    if not "cohort" in byc["response_entity_id"]:
        return r_s_res

    cohorts = []

    # TODO: expand parameters by adding more pre-computed to collations (individual counts ...)

    for cs_i, cs_r in enumerate(r_s_res):
        cohorts.append({
            "cohort_id": cs_r.get("id", cs_i),
            "cohort_type": "beacon-defined",
            "cohort_name": cs_r.get("label", ""),
            "cohort_size": int(cs_r.get("count", 0))
        })

    return cohorts


################################################################################

def remap_runs(r_s_res, byc):
    if not "run" in byc["response_entity_id"]:
        return r_s_res

    runs = []
    for cs_i, cs_r in enumerate(r_s_res):
        r = {
            "id": cs_r.get("id", ""),
            "analysis_id": cs_r.get("id", ""),
            "biosample_id": cs_r.get("biosample_id", ""),
            "individual_id": cs_r.get("individual_id", ""),
            "run_date": datetime.datetime.fromisoformat(
                cs_r.get("updated", datetime.datetime.now().isoformat())).isoformat()
        }
        runs.append(r)

    return runs


################################################################################

def remap_biosamples(r_s_res, byc):
    if not "biosample" in byc["response_entity_id"]:
        return r_s_res

    bs_pop_keys = ["_id", "followup_state", "followup_time"]  # "info"

    for bs_i, bs_r in enumerate(r_s_res):

        # TODO: REMOVE VERIFIER HACKS
        r_s_res[bs_i].update({"sample_origin_type": {"id": "OBI:0001479", "label": "specimen from organism"}})

        for f in ["tumor_grade", "pathological_stage", "histological_diagnosis"]:
            try:
                if f in r_s_res[bs_i]:
                    if not r_s_res[bs_i][f]:
                        r_s_res[bs_i].pop(f, None)
                    elif not r_s_res[bs_i][f]["id"]:
                        r_s_res[bs_i].pop(f, None)
                    elif len(r_s_res[bs_i][f]["id"]) < 4:
                        r_s_res[bs_i].pop(f, None)
            except:
                pass

        for k in bs_pop_keys:
            r_s_res[bs_i].pop(k, None)

    return r_s_res


################################################################################

def remap_individuals(r_s_res, byc):
    if not "individual" in byc["response_entity_id"]:
        return r_s_res

    return r_s_res


################################################################################

def remap_phenopackets(ds_id, r_s_res, byc):
    if not "phenopacket" in byc["response_entity_id"]:
        return r_s_res

    mongo_client = MongoClient()
    data_db = mongo_client[ds_id]
    pxf_s = []

    for ind_i, ind in enumerate(r_s_res):
        pxf = phenopack_individual(ind, data_db, byc)
        pxf_s.append(pxf)

    return pxf_s


################################################################################

def phenopack_individual(ind, data_db, byc):
    # TODO: key removal based on the ones not part of the respective PXF schemas
    # or better on filling them in only for existing parameters

    pxf_bios = []
    ind_pop_keys = ["_id", "provenance", "external_references", "description", "info"]
    bs_pop_keys = ["info", "provenance", "_id", "followup_time", "followup_state", "cohorts", "icdo_morphology",
                   "icdo_topography"]

    pxf_resources = _phenopack_resources(byc)
    server = select_this_server(byc)

    bios_s = data_db["biosamples"].find({"individual_id": ind["id"]})

    for bios in bios_s:

        bios.update({
            "files": [
                {
                    "uri": "{}/beacon/biosamples/{}/variants/?output=vcf".format(server, bios["id"]),
                    "file_attributes": {
                        "genomeAssembly": "GRCh38",
                        "fileFormat": "VCF"
                    }
                }
            ]
        })
        for k in bs_pop_keys:
            bios.pop(k, None)

        clean_empty_fields(bios)

        pxf_bios.append(bios)

    for k in ind_pop_keys:
        ind.pop(k, None)

    individual_remap_pgx_diseases(ind)

    for d_i, d in enumerate(ind["diseases"]):
        for k in ["followup_state", "followup_time"]:
            ind["diseases"][d_i].pop(k, None)

    pxf = {
        "id": re.sub("pgxind", "pgxpxf", ind["id"]),
        "subject": ind,
        "biosamples": pxf_bios,
        "metaData": {
            "submitted_by": "@mbaudis",
            "phenopacket_schema_version": "v2",
            "resources": pxf_resources
        }
    }

    return pxf


################################################################################

def clean_empty_fields(this_object):
    if not isinstance(this_object, dict):
        return this_object

    for k in list(this_object.keys()):
        if isinstance(this_object[k], dict):
            if not this_object[k]:
                this_object.pop(k, None)
        elif isinstance(this_object[k], list):
            if len(this_object[k]) < 1:
                this_object.pop(k, None)

    return this_object


################################################################################

def individual_remap_pgx_diseases(ind):
    # TODO: This should be more general, i.e. what is mapped/how, deleted...

    diseases = []

    for k in ["index_disease", "auxiliary_disease"]:
        try:
            if len(ind[k]["disease_code"]["id"]) > 1:
                diseases.append(ind[k])
        except:
            pass
        ind.pop(k, None)

    ind.update({"diseases": diseases})
    return ind


################################################################################

def _phenopack_resources(byc):
    # TODO: make this general, at least for phenopacket response, and only scan used prefixes

    f_d = byc["filter_definitions"]
    # rkeys = ["NCITgrade", "NCITstage", "NCITtnm", "NCIT", "PATOsex", "EFOfus" ]
    rkeys = ["NCIT", "PATOsex", "EFOfus", "UBERON"]

    pxf_rs = []

    for r_k in rkeys:
        if r_k not in f_d.keys():
            continue

        d = f_d[r_k]

        pxf_r = {
            "id": r_k,
            "name": d.get("name", ""),
            "url": d.get("url", ""),
            "namespace_prefix": d.get("namespace_prefix", ""),
            "version": "2022-04-01",
            "iri_prefix": "http://purl.obolibrary.org/obo/{}_".format(d.get("namespace_prefix", ""))
        }

        pxf_rs.append(pxf_r)

    return pxf_rs


################################################################################

def remap_all(r_s_res):
    for br_i, br_r in enumerate(r_s_res):
        if type(br_r) is not dict:
            continue
        r_s_res[br_i].pop("_id", None)
        clean_empty_fields(r_s_res[br_i])

    return r_s_res


################################################################################

def normalize_variant_values_for_export(v, byc, drop_fields=None):
    drop_fields = [] if drop_fields is None else drop_fields

    v_defs = byc["variant_definitions"]

    v["log2"] = False
    if "info" in v:
        if "cnv_value" in v["info"]:
            if isinstance(v["info"]["cnv_value"], float):
                v["log2"] = round(v["info"]["cnv_value"], 3)
        if not "info" in drop_fields:
            drop_fields.append("info")

    if v["log2"] == False:
        if "variant_type" in v:
            if v["variant_type"] in v_defs["cnv_dummy_values"].keys():
                v["log2"] = v_defs["cnv_dummy_values"][v["variant_type"]]

    if v["log2"] == False:
        drop_fields.append("log2")

    for i_k in ["start", "end"]:
        v.update({i_k: int(v[i_k])})

    for d_f in drop_fields:
        v.pop(d_f, None)

    return v


################################################################################

def callset__ids_create_iset(ds_id, label, cs__ids, byc):
    mongo_client = MongoClient()
    cs_coll = mongo_client[ds_id]["callsets"]

    cs_cursor = cs_coll.find({"_id": {"$in": cs__ids}, "variant_class": {"$ne": "SNV"}})
    intervals, cnv_cs_count = interval_counts_from_callsets(cs_cursor, byc)

    iset = {
        "dataset_id": ds_id,
        "group_id": ds_id,
        "label": label,
        "sample_count": cnv_cs_count,
        "interval_frequencies": []
    }

    for intv_i, intv in enumerate(intervals):
        iset["interval_frequencies"].append(intv.copy())

    return iset


################################################################################

def callsets_create_iset(ds_id, label, callsets, byc):
    intervals, cnv_cs_count = interval_counts_from_callsets(callsets, byc)

    iset = {
        "dataset_id": ds_id,
        "group_id": ds_id,
        "label": label,
        "sample_count": cnv_cs_count,
        "interval_frequencies": []
    }

    for intv_i, intv in enumerate(intervals):
        iset["interval_frequencies"].append(intv.copy())

    return iset



import datetime
import io
import re
import base64
from PIL import Image, ImageDraw

from cgi_parsing import get_plot_parameters, print_svg_response, prjsonnice, test_truthy
from clustering_utils import cluster_frequencies, cluster_samples
from cytoband_utils import bands_from_cytobands, retrieve_gene_id_coordinates

# http://progenetix.org/cgi/bycon/services/intervalFrequencies.py?chr2plot=8,9,17&labels=8:120000000-123000000:Some+Interesting+Region&plot_gene_symbols=MYCN,REL,TP53,MTAP,CDKN2A,MYC,ERBB2,CDK1&filters=pgx:icdom-85003&output=histoplot
# http://progenetix.org/beacon/biosamples/?datasetIds=progenetix&referenceName=9&variantType=DEL&start=21500000&start=21975098&end=21967753&end=22500000&filters=NCIT:C3058&output=histoplot&plotGeneSymbols=CDKN2A,MTAP,EGFR,BCL6
# http://progenetix.org/beacon/biosamples/?datasetIds=progenetix&referenceName=9&variantType=DEL&start=21500000&start=21975098&end=21967753&end=22500000&filters=NCIT:C3058&output=samplesplot&plotGeneSymbols=CDKN2A,MTAP,EGFR,BCL6

################################################################################
################################################################################
################################################################################

class ByconPlot:
    """
    # The `ByconPlot` class

    ## Input

    A plot data bundle containing lists of callset object bundles (_i.e._ the
    callsets with all their individual variants added) and/or interval frequencies
    set bundles (_i.e._ list of one or more binned CNV frequencies in object
    wrappers with some information about the set).

    """

    def __init__(self, byc: dict, plot_data_bundle: dict):

        self.env = byc.get("env", "server")
        self.byc = byc
        self.plot_data_bundle = plot_data_bundle
        self.svg = None

    # -------------------------------------------------------------------------#
    # ----------------------------- public ------------------------------------#
    # -------------------------------------------------------------------------#

    def get_svg(self) -> str:
        self.__plot_pipeline()
        return self.svg

    def svg2file(self, filename):
        self.__plot_pipeline()
        svg_fh = open(filename, "w")
        svg_fh.write(self.svg)
        svg_fh.close()

    def svg_response(self):
        self.__plot_pipeline()
        print_svg_response(self.svg, self.env)

    # -------------------------------------------------------------------------#
    # ----------------------------- testing -----------------------------------#
    # -------------------------------------------------------------------------#

    def __plot_add_probesplot(self):
        """
        Prototyping bitmap drawing for probe plots etc.
        Invoked w/ &output=arrayplot
        https://pillow.readthedocs.io/en/stable/reference/ImageDraw.html
        
        #### Draw examples

        * draw.point((50,50), (50,255,0))
        * draw.line((0, 0) + image.size, fill=128)
        * draw.line((0, image.size[1], image.size[0], 0), fill=(50,255,0))
        * draw.rectangle([0, 0, 28, image.size[1]], fill="rgb(255,20,66)")
        * draw.ellipse([(80,20),(130,50)], fill="#ccccff", outline="red")
        
        #### Input:
        ```
        probes = [
            {
                "reference_name": "17",
                "start": 13663925,
                "value": 2.5
            },
            {...}
        ]
        ```
        """

        if not "samplesplot" in self.plv["plot_type"]:
            return

        p_t_s = self.byc["plot_defaults"]["plot_types"]
        d_k = p_t_s["samplesplot"].get("data_key")

        probebundles = self.plot_data_bundle.get(d_k, [{"id":"___undefined___"}])
        if len(probebundles) != 1:
            return
        if not "cn_probes" in probebundles[0]:
            return

        probes = probebundles[0].get("cn_probes", [])
        # self.plv["plot_area_height"] * 0.5 / self.plv["plot_axis_y_max"]
        self.plv.update({
            "plot_axis_y_max": 4,
            "plot_y2pf": self.plv["plot_area_height"] * 0.5 / 4 * self.plv["plot_probe_y_factor"],
            "plot_first_area_y0": self.plv["Y"],
            "plot_label_y_unit": "",
            "plot_label_y_values": self.plv["plot_probe_label_y_values"]
        })

        X = 0
        h_y_0 = self.plv["plot_area_height"] * 0.5
        p_y_f = self.plv["plot_y2pf"]
        p_half = self.plv["plot_probedot_size"] * 0.5
        p_dense = self.plv["plot_probedot_opacity"]

        if len(probes) > 500000:
            p_half *= 0.5
            p_dense = p_dense * 0.7
        p_dense = int(round(p_dense, 0))

        image = Image.new(
                    'RGBA',
                    (self.plv["plot_area_width"], self.plv["plot_area_height"]),
                    color=self.plv["plot_area_color"]
                )
        draw = ImageDraw.Draw(image)

        for chro in self.plv["plot_chros"]:

            c_p = list(filter(lambda d: d["reference_name"] == chro, probes.copy()))
            c_l = self.byc["cytolimits"][str(chro)]
            chr_w = c_l["size"] * self.plv["plot_b2pf"]

            i_x_0 = X

            for i_v in c_p:
                s = i_x_0 + i_v.get("start", 0) * self.plv["plot_b2pf"]
                v = i_v.get("value", 0)
                h = v * p_y_f 
                if h > h_y_0:
                    h = h_y_0
                if h < -h_y_0:
                    h = -h_y_0
                h_p = h_y_0 - h

                # draw.ellipse(
                #     [
                #         (s-p_half, h_p - p_half),
                #         (s+p_half, h_p + p_half)
                #     ],
                #     fill=(0,0,63,p_dense)
                # )
                draw.point((round(s, 2),round(h_p, 2)), (0,0,63,p_dense))

            X += chr_w + self.plv["plot_region_gap_width"]

        # ------------------------ / histogram data ----------------------------#

        in_mem_file = io.BytesIO()
        image.save(in_mem_file, format = "PNG")
        in_mem_file.seek(0)
        img_bytes = in_mem_file.read()
        base64_encoded_result_bytes = base64.b64encode(img_bytes)
        base64_encoded_result_str = base64_encoded_result_bytes.decode('ascii')

        self.plv["pls"].append("""
<image
  x="{}"
  y="{}"
  width="{}"
  height="{}"
  xlink:href="data:image/png;base64,{}"
/>""".format(
            self.plv["plot_area_x0"],
            self.plv["Y"],
            self.plv["plot_area_width"],
            self.plv["plot_area_height"],
            base64_encoded_result_str
        ))

        self.__plot_area_add_grid()

        self.plv["Y"] += self.plv["plot_area_height"]
        self.plv.update({"plot_last_area_ye": self.plv["Y"]})
        self.plv["Y"] += self.plv["plot_region_gap_width"]

    # -------------------------------------------------------------------------#
    # ----------------------------- private -----------------------------------#
    # -------------------------------------------------------------------------#

    def __plot_pipeline(self):

        p_t_s = self.byc["plot_defaults"].get("plot_types", {})
        p_t = self.byc.get("output", "___none___")
        if p_t not in p_t_s.keys():
            return

        self.__initialize_plot_values()

        if self.__plot_respond_empty_results() is False:
            self.__plot_add_title()
            self.__plot_add_cytobands()
            self.__plot_add_samplestrips()
            self.__plot_add_histodata()
            self.__plot_add_probesplot()
            self.__plot_add_cluster_tree()
            self.__plot_add_markers()

        self.__plot_add_footer()

        self.svg = self.__create_svg()

    # -------------------------------------------------------------------------#

    def __initialize_plot_values(self):
        p_d_p = self.byc["plot_defaults"]["parameters"]
        p_t_s = self.byc["plot_defaults"]["plot_types"]
        p_t = self.byc["output"]

        d_k = p_t_s[p_t].get("data_key")

        # TODO: get rid of the "results"?
        self.plv = {
            "plot_type": p_t,
            "results": self.plot_data_bundle.get(d_k, []),
            "results_number": len(self.plot_data_bundle.get(d_k, []))
        }

        self.__filter_empty_callsets_results()

        for p_k, p_d in p_d_p.items():
            if "default" in p_d:
                self.plv.update({p_k: p_d["default"]})
            else:
                self.plv.update({p_k: ""})

        if self.plv["results_number"] < 2:
            self.plv.update({"plot_labelcol_width": 0})

        if self.plv["results_number"] > 2:
            self.plv.update({"plot_cluster_results": True})
        else:
            self.plv.update({"plot_dendrogram_width": 0})

        get_plot_parameters(self.plv, self.byc)

        pax = self.plv["plot_margins"] + self.plv["plot_labelcol_width"] + self.plv["plot_axislab_y_width"]

        paw = self.plv["plot_width"] - 2 * self.plv["plot_margins"]
        paw -= self.plv["plot_labelcol_width"]
        paw -= self.plv["plot_axislab_y_width"]
        paw -= self.plv["plot_dendrogram_width"]

        # calculate the base
        chr_b_s = 0
        for chro in self.plv["plot_chros"]:
            c_l = self.byc["cytolimits"][chro]
            chr_b_s += c_l["size"]

        pyf = self.plv["plot_area_height"] * 0.5 / self.plv["plot_axis_y_max"]

        gaps = len(self.plv["plot_chros"]) - 1
        gap_sw = gaps * self.plv["plot_region_gap_width"]
        genome_width = paw - gap_sw
        b2pf = genome_width / chr_b_s  # TODO: only exists if using stack

        title = self.plv.get("plot_title", "")
        if len(title) < 3:
            if self.plv["results_number"] == 1:
                title = self.__format_resultset_title()

        self.plv.update({
            "plot_title": title,
            "cytoband_shades": self.byc["plot_defaults"].get("cytoband_shades", {}),
            "styles": [
                f'.plot-area {{fill: {self.plv.get("plot_area_color", "#66ddff")}; fill-opacity: {self.plv.get("plot_area_opacity", 0.8)};}}'],
            "Y": self.plv["plot_margins"],
            "plot_area_width": paw,
            "plot_area_x0": pax,
            "plot_area_xe": pax + paw,
            "plot_area_xc": pax + paw / 2,
            "plot_y2pf": pyf,
            "plot_genome_size": chr_b_s,
            "plot_b2pf": b2pf,
            "plot_labels": {},
            "dendrogram": False,
            "pls": []
        })

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __filter_empty_callsets_results(self):
        if not "samplesplot" in self.plv["plot_type"]:
            return

        p_t_s = self.byc["plot_defaults"]["plot_types"]
        d_k = p_t_s["samplesplot"].get("data_key")

        if test_truthy(self.plv.get("plot_filter_empty_samples", False)):
            self.plot_data_bundle.update({d_k: [s for s in self.plot_data_bundle[d_k] if len(s['variants']) > 0]})

        self.plv.update({
            "results": self.plot_data_bundle[d_k],
            "results_number": len(self.plot_data_bundle[d_k])
        })

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_respond_empty_results(self):
        if self.plv["results_number"] > 0:
            return False

        self.plv.update({
            "plot_title_font_size": self.plv["plot_font_size"],
            "plot_title": "No matching CNV data"
        })

        self.__plot_add_title()

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __format_resultset_title(self):

        title = ""

        f_set = self.plv["results"][0]

        g_id = f_set.get("group_id")
        g_lab = f_set.get("label")
        if g_lab is not None:
            title = f"{g_lab}"
            if g_id is not None:
                title += f" ({g_id})"
        elif g_id is not None:
            title = g_id

        return title

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_add_title(self):

        if len(self.plv.get("plot_title", "")) < 3:
            return

        self.plv["Y"] += self.plv["plot_title_font_size"]

        self.plv["pls"].append(
            '<text x="{}" y="{}" style="text-anchor: middle; font-size: {}px">{}</text>'.format(
                self.plv["plot_area_xc"],
                self.plv["Y"],
                self.plv["plot_title_font_size"],
                self.plv["plot_title"]
            )
        )

        self.plv["Y"] += self.plv["plot_title_font_size"]

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_add_cytobands(self):

        if self.plv["plot_chro_height"] < 1:
            return

        self.__plot_add_cytoband_svg_gradients()

        # ------------------------- chromosome labels --------------------------#

        x = self.plv["plot_area_x0"]
        self.plv["Y"] += self.plv["plot_title_font_size"]

        for chro in self.plv["plot_chros"]:
            c_l = self.byc["cytolimits"][str(chro)]

            chr_w = c_l["size"] * self.plv["plot_b2pf"]
            chr_c = x + chr_w / 2

            self.plv["pls"].append(
                f'<text x="{chr_c}" y="{self.plv["Y"]}" style="text-anchor: middle; font-size: {self.plv["plot_font_size"]}px">{chro}</text>')

            x += chr_w
            x += self.plv["plot_region_gap_width"]

        self.plv["Y"] += self.plv["plot_region_gap_width"]

        # ---------------------------- chromosomes ----------------------------#

        x = self.plv["plot_area_x0"]
        self.plv.update({"plot_chromosomes_y0": self.plv["Y"]})

        for chro in self.plv["plot_chros"]:

            c_l = self.byc["cytolimits"][str(chro)]
            chr_w = c_l["size"] * self.plv["plot_b2pf"]

            chr_cb_s = list(filter(lambda d: d["chro"] == chro, self.byc["cytobands"].copy()))

            last = len(chr_cb_s) - 1
            this_n = 0

            for cb in chr_cb_s:

                this_n += 1

                s_b = cb["start"]
                e_b = cb["end"]
                c = cb["staining"]
                cb_l = int(e_b) - int(s_b)
                l_px = cb_l * self.plv["plot_b2pf"]

                by = self.plv["Y"]
                bh = self.plv["plot_chro_height"]

                if "cen" in c:
                    by += 0.2 * self.plv["plot_chro_height"]
                    bh -= 0.4 * self.plv["plot_chro_height"]
                elif "stalk" in c:
                    by += 0.3 * self.plv["plot_chro_height"]
                    bh -= 0.6 * self.plv["plot_chro_height"]
                elif this_n == 1 or this_n == last:
                    by += 0.1 * self.plv["plot_chro_height"]
                    bh -= 0.2 * self.plv["plot_chro_height"]

                self.plv["pls"].append(
                    f'<rect x="{round(x, 1)}" y="{round(by, 1)}" width="{round(l_px, 1)}" height="{round(bh, 1)}" style="fill: url(#{self.plv["plot_id"]}{c}); " />')

                x += l_px

            x += self.plv["plot_region_gap_width"]

        # -------------------------- / chromosomes -----------------------------#

        self.plv["Y"] += self.plv["plot_chro_height"]
        self.plv["Y"] += self.plv["plot_region_gap_width"]

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_add_cytoband_svg_gradients(self):

        c_defs = ""

        for cs_k, cs_c in self.plv["cytoband_shades"].items():

            p_id = self.plv.get("plot_id", "")
            c_defs += f'\n<linearGradient id="{p_id}{cs_k}" x1="0%" x2="0%" y1="0%" y2="80%" spreadMethod="reflect">'

            for k, v in cs_c.items():
                c_defs += f'\n  <stop offset="{k}" stop-color="{v}" />'

            c_defs += f'\n</linearGradient>'

        self.plv["pls"].insert(0, c_defs)

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_add_samplestrips(self):

        if not "sample" in self.plv["plot_type"]:
            return

        self.plv.update({"plot_first_area_y0": self.plv["Y"]})
        self.plv["pls"].append("")
        self.plv.update({"plot_strip_bg_i": len(self.plv["pls"]) - 1})

        lab_x_e = self.plv["plot_area_x0"] - self.plv["plot_region_gap_width"] * 2
        lab_f_s = round(self.plv["plot_samplestrip_height"] * 0.65, 1)

        if len(self.plv["results"]) > 0:

            self.__plot_order_samples()

            if len(self.plv["results"]) > 1:
                self.plv["styles"].append(
                    f'.title-left {{text-anchor: end; fill: {self.plv["plot_font_color"]}; font-size: {lab_f_s}px;}}'
                )

            for s in self.plv["results"]:
                self.__plot_add_one_samplestrip(s)
                if lab_f_s > 5 and len(self.plv["results"]) > 1:
                    cs_id = s.get("callset_id", "")
                    if len(cs_id) > 0:
                        cs_id = f' ({cs_id})'
                    g_lab = f'{s.get("biosample_id", "")}{cs_id}'
                    self.plv["pls"].append(
                        f'<text x="{lab_x_e}" y="{self.plv["Y"] - round(self.plv["plot_samplestrip_height"] * 0.2, 1)}" class="title-left">{g_lab}</text>')

        self.plv["plot_last_area_ye"] = self.plv["Y"]

        # ----------------------- plot cluster tree --------------------------------#

        self.plv.update({"cluster_head_gap": 0})
        self.plv.update({"plot_clusteritem_height": self.plv["plot_samplestrip_height"]})

        # --------------------- plot area background -------------------------------#

        x_a_0 = self.plv["plot_area_x0"]
        p_a_w = self.plv["plot_area_width"]
        p_a_h = self.plv["Y"] - self.plv["plot_first_area_y0"]

        self.plv["pls"][self.plv[
            "plot_strip_bg_i"]] = f'<rect x="{x_a_0}" y="{self.plv["plot_first_area_y0"]}" width="{p_a_w}" height="{p_a_h}" class="plot-area" />'
        self.plv["Y"] += self.plv["plot_region_gap_width"]

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_order_samples(self):

        if self.plv.get("plot_cluster_results", True) is True and len(self.plv["results"]) > 2:
            dendrogram = cluster_samples(self.plv, self.byc)
            new_order = dendrogram.get("leaves", [])
            if len(new_order) == len(self.plv["results"]):
                self.plv["results"][:] = [self.plv["results"][i] for i in dendrogram.get("leaves", [])]
                self.plv.update({"dendrogram": dendrogram})

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_add_one_samplestrip(self, s):

        v_s = s.get("variants", [])

        x = self.plv["plot_area_x0"]
        h = self.plv["plot_samplestrip_height"]

        cnv_c = {
            "DUP": self.plv["plot_dup_color"],
            "DEL": self.plv["plot_del_color"]
        }

        for chro in self.plv["plot_chros"]:

            c_l = self.byc["cytolimits"][str(chro)]
            chr_w = c_l["size"] * self.plv["plot_b2pf"]

            c_v_s = list(filter(lambda d: d["reference_name"] == chro, v_s.copy()))

            for p_v in c_v_s:
                s_v = int(p_v.get("start", 0))
                e_v = int(p_v.get("end", s_v))
                l_v = round((e_v - s_v) * self.plv["plot_b2pf"], 1)
                if l_v < 0.5:
                    l_v = 0.5
                v_x = round(x + s_v * self.plv["plot_b2pf"], 1)
                t = p_v.get("variant_type", "NA")
                c = cnv_c.get(t, "rgb(111,111,111)")

                self.plv["pls"].append(
                    f'<rect x="{v_x}" y="{self.plv["Y"]}" width="{l_v}" height="{h}" style="fill: {c} " />')

            x += chr_w
            x += self.plv["plot_region_gap_width"]

        self.plv["Y"] += h

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_add_cluster_tree(self):

        itemHeight = self.plv["plot_clusteritem_height"]

        d = self.plv.get("dendrogram", False)

        if d is False:
            return

        p_s_c = self.plv.get("plot_dendrogram_color", '#ee0000')
        p_s_w = self.plv.get("plot_dendrogram_stroke", 1)

        d_x_s = d.get("dcoord", [])
        d_y_s = d.get("icoord", [])

        t_y_0 = self.plv["plot_first_area_y0"]
        t_x_0 = self.plv["plot_area_x0"] + self.plv["plot_area_width"]
        t_y_f = itemHeight * 0.1

        # finding the largest x-value of the dendrogram for scaling
        x_max = self.plv["plot_dendrogram_width"]

        for i, node in enumerate(d_x_s):
            for j, x in enumerate(node):
                if x > x_max:
                    x_max = x
        t_x_f = self.plv["plot_dendrogram_width"] / x_max

        for i, node in enumerate(d_x_s):

            n = f'<polyline points="'

            for j, x in enumerate(node):
                y = d_y_s[i][j] * t_y_f - self.plv["cluster_head_gap"]

                for h, f_set in enumerate(self.plv["results"]):
                    h_y_e = h * (itemHeight + self.plv["cluster_head_gap"])
                    if y > h_y_e:
                        y += self.plv["cluster_head_gap"]

                n += f' {round(t_x_0 + x * t_x_f, 1)},{round(t_y_0 + y, 1)}'

            n += f'" fill="none" stroke="{p_s_c}" stroke-width="{p_s_w}px" />'

            self.plv["pls"].append(n)

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_add_histodata(self):

        if "histo" not in self.plv["plot_type"]:
            return

        self.plv.update({"plot_first_area_y0": self.plv["Y"]})

        self.__plot_order_histograms()

        for f_set in self.plv["results"]:
            self.__plot_add_one_histogram(f_set)

        self.plv["plot_last_area_ye"] = self.plv["Y"]

        # ----------------------- plot cluster tree ----------------------------#

        self.plv.update({"cluster_head_gap": self.plv["plot_region_gap_width"]})
        self.plv.update({"plot_clusteritem_height": self.plv["plot_area_height"]})

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_order_histograms(self):

        if self.plv.get("plot_cluster_results", True) is True and len(self.plv["results"]) > 2:
            dendrogram = cluster_frequencies(self.plv, self.byc)
            new_order = dendrogram.get("leaves", [])
            if len(new_order) == len(self.plv["results"]):
                self.plv["results"][:] = [self.plv["results"][i] for i in dendrogram.get("leaves", [])]
                self.plv.update({"dendrogram": dendrogram})

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_add_one_histogram(self, f_set):

        self.__plot_add_one_histogram_canvas(f_set)

        i_f = f_set.get("interval_frequencies", [])

        X = self.plv["plot_area_x0"]
        h_y_0 = self.plv["Y"] + self.plv["plot_area_height"] * 0.5

        # ------------------------- histogram data -----------------------------#

        # TODO: in contrast to the Perl version here we don't correct for interval
        #       sets which _do not_ correspond to the full chromosome coordinates

        cnv_c = {
            "gain_frequency": self.plv["plot_dup_color"],
            "loss_frequency": self.plv["plot_del_color"]
        }
        cnv_f = {"gain_frequency": -1, "loss_frequency": 1}

        for chro in self.plv["plot_chros"]:

            c_l = self.byc["cytolimits"][str(chro)]
            chr_w = c_l["size"] * self.plv["plot_b2pf"]

            c_i_f = list(filter(lambda d: d["reference_name"] == chro, i_f.copy()))
            c_i_no = len(c_i_f)

            for GL in ["gain_frequency", "loss_frequency"]:

                p_c = cnv_c[GL]
                h_f = cnv_f[GL]

                p = f'<polygon points="{round(X, 1)},{round(h_y_0, 1)}'

                i_x_0 = X
                prev = -1
                for c_i_i, i_v in enumerate(c_i_f, start=1):

                    s = i_x_0 + i_v.get("start", 0) * self.plv["plot_b2pf"]
                    e = i_x_0 + i_v.get("end", 0) * self.plv["plot_b2pf"]
                    v = i_v.get(GL, 0)
                    h = v * self.plv["plot_y2pf"]
                    h_p = h_y_0 + h * h_f

                    point = f' {round(s, 1)},{round(h_p, 1)}'

                    # This construct avoids adding intermediary points w/ the same
                    # value as the one before and after 
                    if c_i_no > c_i_i:
                        future = c_i_f[c_i_i].get(GL, 0)
                        if prev != v or future != v:
                            p += point
                    else:
                        p += point

                    prev = v

                p += f' {round((X + chr_w), 1)},{round(h_y_0, 1)}" fill="{p_c}" stroke-width="0px" />'
                self.plv["pls"].append(p)

            X += chr_w
            X += self.plv["plot_region_gap_width"]

        # ------------------------ / histogram data ----------------------------#

        self.plv["Y"] += self.plv["plot_area_height"]
        self.plv.update({"plot_last_area_ye": self.plv["Y"]})
        self.plv["Y"] += self.plv["plot_region_gap_width"]

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_add_one_histogram_canvas(self, f_set):

        x_a_0 = self.plv["plot_area_x0"]
        p_a_w = self.plv["plot_area_width"]
        p_a_h = self.plv["plot_area_height"]

        # -------------------------- left labels -------------------------------#

        self.__histoplot_add_left_label(f_set)

        # --------------------- plot area background ---------------------------#

        self.plv["pls"].append(
            f'<rect x="{x_a_0}" y="{self.plv["Y"]}" width="{p_a_w}" height="{p_a_h}" class="plot-area" />')

        # --------------------------- grid lines -------------------------------#

        self.__plot_area_add_grid()

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __histoplot_add_left_label(self, f_set):

        if self.plv["plot_labelcol_width"] < 10:
            return

        lab_x_e = self.plv["plot_margins"] + self.plv["plot_labelcol_width"]
        h_y_0 = self.plv["Y"] + self.plv["plot_area_height"] * 0.5

        self.plv["styles"].append(
            f'.title-left {{text-anchor: end; fill: {self.plv["plot_font_color"]}; font-size: {self.plv["plot_labelcol_font_size"]}px;}}'
        )

        g_id = f_set.get("group_id", "NA")
        g_ds_id = f_set.get("dataset_id", False)
        g_lab = f_set.get("label", "")
        g_no = f_set.get("sample_count", 0)

        # The condition splits the label data on 2 lines if a text label pre-exists
        if len(self.byc["dataset_ids"]) > 1 and g_ds_id is not False:
            count_lab = f' ({g_ds_id}, {g_no} {"samples" if g_no > 1 else "sample"})'
        else:
            count_lab = f' ({g_no} {"samples" if g_no > 1 else "sample"} )'

        if len(g_lab) > 0:
            lab_y = h_y_0 - self.plv["plot_labelcol_font_size"] * 0.2
            self.plv["pls"].append(f'<text x="{lab_x_e}" y="{lab_y}" class="title-left">{g_lab}</text>')
            lab_y = h_y_0 + self.plv["plot_labelcol_font_size"] * 1.2
            self.plv["pls"].append(f'<text x="{lab_x_e}" y="{lab_y}" class="title-left">{g_id}{count_lab}</text>')
        else:
            lab_y = h_y_0 - self.plv["plot_labelcol_font_size"] * 0.5
            self.plv["pls"].append(f'<text x="{lab_x_e}" y="{lab_y}" class="title-left">{g_id}{count_lab}</text>')

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_area_add_grid(self):

        x_a_0 = self.plv["plot_area_x0"]
        x_c_e = self.plv["plot_area_xe"]

        h_y_0 = self.plv["Y"] + self.plv["plot_area_height"] * 0.5
        x_y_l = x_a_0 - self.plv["plot_region_gap_width"]

        u = self.plv["plot_label_y_unit"]
 
        self.plv["styles"].append(
            f'.label-y {{text-anchor: end; fill: {self.plv["plot_label_y_font_color"]}; font-size: {self.plv["plot_label_y_font_size"]}px;}}'
        )
        self.plv["styles"].append(
            f'.gridline {{stroke-width: {self.plv["plot_grid_stroke"]}px; stroke: {self.plv["plot_grid_color"]}; opacity: {self.plv["plot_grid_opacity"]} ; }}',
        )

        # -------------------------- center line -----------------------------------#

        self.plv["pls"].append(
            f'<line x1="{x_a_0 - self.plv["plot_region_gap_width"]}"  y1="{h_y_0}"  x2="{x_c_e}"  y2="{h_y_0}" class="gridline" />')

        # --------------------------- grid lines -----------------------------------#

        for y_m in self.plv["plot_label_y_values"]:

            if y_m >= self.plv["plot_axis_y_max"]:
                continue

            for f in [1, -1]:
                if u == "" and f == 1:
                    neg = "-"
                else:
                    neg = ""

                y_v = h_y_0 + f * y_m * self.plv["plot_y2pf"]
                y_l_y = y_v + self.plv["plot_label_y_font_size"] / 2

                self.plv["pls"].append(f'<line x1="{x_a_0}" y1="{y_v}" x2="{x_c_e}" y2="{y_v}" class="gridline" />')

                if self.plv["plot_axislab_y_width"] < 1:
                    continue

                self.plv["pls"].append(f'<text x="{x_y_l}" y="{y_l_y}" class="label-y">{neg}{y_m}{u}</text>')

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_add_markers(self):

        self.__add_labs_from_plot_region_labels()
        self.__add_labs_from_gene_symbols()
        self.__add_labs_from_cytobands()

        labs = self.plv.get("plot_labels", [])

        if len(labs) < 1:
            return

        b2pf = self.plv["plot_b2pf"]

        p_m_f_s = self.plv["plot_marker_font_size"]
        p_m_l_p = self.plv["plot_marker_label_padding"]
        p_m_lane_p = self.plv["plot_marker_lane_padding"]
        p_m_l_h = p_m_f_s + p_m_l_p * 2
        p_m_lane_h = p_m_l_h + p_m_lane_p

        max_lane = 0
        marker_y_0 = round(self.plv["plot_first_area_y0"], 1)
        marker_y_e = round(self.plv["plot_last_area_ye"] + p_m_lane_p, 1)

        x = self.plv["plot_area_x0"]

        m_p_e = [(x - 30)]
        for chro in self.plv["plot_chros"]:

            c_l = self.byc["cytolimits"][chro]
            chr_w = c_l["size"] * self.plv["plot_b2pf"]

            for m_k, m_v in labs.items():

                c = str(m_v.get("chro", "__na__"))

                if str(chro) != c:
                    continue

                s = int(m_v.get("start", 0))
                e = int(m_v.get("end", 0))
                label = m_v.get("label", "")

                m_s = x + s * b2pf
                m_e = x + e * b2pf
                m_w = m_e - m_s
                if 1 > m_w > 0:
                    m_w = 1
                else:
                    m_w = round(m_w, 1)
                m_c = round((m_s + m_e) / 2, 1)
                m_l_w = len(label) * 0.75 * p_m_f_s
                m_l_s = m_c - 0.5 * m_l_w
                m_l_e = m_c + 0.5 * m_l_w

                found_space = False
                l_i = 0

                for p_e in m_p_e:
                    if m_l_s > p_e:
                        found_space = True
                        m_p_e[l_i] = m_l_e
                        break
                    l_i += 1

                if found_space is False:
                    m_p_e.append(m_l_e)

                if len(m_p_e) > max_lane:
                    max_lane = len(m_p_e)

                m_y_e = marker_y_e + l_i * p_m_lane_h
                m_h = round(m_y_e - marker_y_0, 1)
                l_y_p = marker_y_e + l_i * p_m_lane_h + p_m_lane_h - p_m_l_p - p_m_lane_p - 1

                self.plv["pls"].append(
                    f'<rect x="{round(m_s, 1)}" y="{marker_y_0}" width="{round(m_w, 1)}" height="{m_h}" class="marker" style="opacity: 0.4; " />')
                self.plv["pls"].append(
                    f'<rect x="{round(m_l_s, 1)}" y="{m_y_e}" width="{round(m_l_w, 1)}" height="{p_m_l_h}" class="marker" style="opacity: 0.1; " />')
                self.plv["pls"].append(f'<text x="{m_c}" y="{l_y_p}" class="marker">{label}</text>')

            x += chr_w
            x += self.plv["plot_region_gap_width"]

        # --------------------- end chromosome loop --------------------------------#

        if max_lane > 0:
            self.plv["Y"] += max_lane * p_m_lane_h
            self.plv["Y"] += self.plv["plot_region_gap_width"]
            self.plv["styles"].append(
                f'.marker {{text-anchor: middle; fill: {self.plv["plot_marker_font_color"]}; font-size: {p_m_f_s}px;}}'
            )

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __add_labs_from_plot_region_labels(self):

        r_l_s = self.plv.get("plot_region_labels", [])
        if len(r_l_s) < 1:
            return

        for label in r_l_s:

            l_i = re.split(":", label)
            if len(l_i) < 2:
                continue
            c = l_i.pop(0)
            s_e_i = l_i.pop(0)
            s_e = re.split("-", s_e_i)
            s = s_e.pop(0)
            # TODO: check r'^\d+?$'
            if len(s_e) < 1:
                e = str(int(s) + 1)
            else:
                e = s_e.pop(0)
            if len(l_i) > 0:
                label = str(l_i.pop(0))
            else:
                label = ""

            m = self.__make_marker_object(c, s, e, label)
            self.plv["plot_labels"].update(m)

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __add_labs_from_gene_symbols(self):

        g_s_s = self.plv.get("plot_gene_symbols", [])
        if len(g_s_s) < 1:
            return

        g_l = []

        for q_g in g_s_s:
            genes, e = retrieve_gene_id_coordinates(q_g, self.byc)
            g_l += genes

        for f_g in g_l:

            m = self.__make_marker_object(
                f_g.get("reference_name", False),
                f_g.get("start", False),
                f_g.get("end", False),
                self.plv.get("plot_marker_font_color", "#ccccff"),
                f_g.get("symbol", False)
            )

            if m is not False:
                self.plv["plot_labels"].update(m)

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __add_labs_from_cytobands(self):

        g_s_s = self.plv.get("plot_cytoregion_labels", [])

        if len(g_s_s) < 1:
            return

        g_l = []

        for q_g in g_s_s:
            cytoBands, chro, start, end, error = bands_from_cytobands(q_g, self.byc)

            if len(cytoBands) < 1:
                continue

            m = self.__make_marker_object(
                chro,
                start,
                end,
                self.plv.get("plot_cytoregion_color", "#ccccff"),
                q_g
            )

            if m is not None:
                self.plv["plot_labels"].update(m)

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __make_marker_object(self, chromosome, start, end, color, label=""):

        m = None

        # Checks here or upstream?
        # Cave: `any` ... `is False` to avoid `True` for `0` with `False in`
        if any(x is False for x in [chromosome, start, end, label]):
            return m

        m_k = f'{chromosome}:{start}-{end}:{label}'

        m = {
            m_k: {
                "chro": chromosome,
                "start": start,
                "end": end,
                "label": label,
                "color": color
            }
        }

        return m

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __plot_add_footer(self):

        today = datetime.date.today()
        x_a_0 = self.plv["plot_area_x0"]
        x_c_e = x_a_0 + self.plv["plot_area_width"]

        self.plv["styles"].append(
            f'.footer-r {{text-anchor: end; fill: {self.plv["plot_footer_font_color"]}; font-size: {self.plv["plot_footer_font_size"]}px;}}'
        )
        self.plv["styles"].append(
            f'.footer-l {{text-anchor: start; fill: {self.plv["plot_footer_font_color"]}; font-size: {self.plv["plot_footer_font_size"]}px;}}'
        )

        self.plv["Y"] += self.plv["plot_footer_font_size"]
        self.plv["pls"].append(
            f'<text x="{x_c_e}" y="{self.plv["Y"]}" class="footer-r">&#169; CC-BY 2001 - {today.year} progenetix.org</text>')

        if self.plv.get("results_number", 0) > 1:
            self.plv["pls"].append(
                f'<text x="{x_a_0}" y="{self.plv["Y"]}" class="footer-l">{self.plv["results_number"]} analyses</text>')

        self.plv["Y"] += self.plv["plot_margins"]

    # --------------------------------------------------------------------------#
    # --------------------------------------------------------------------------#

    def __create_svg(self):

        svg = """<svg
xmlns="http://www.w3.org/2000/svg"
xmlns:xlink="http://www.w3.org/1999/xlink"
version="1.1"
id="{}"
width="{}px"
height="{}px"
style="margin: auto; font-family: Helvetica, sans-serif;">
<style type="text/css"><![CDATA[
{}
]]></style>
<rect x="0" y="0" width="{}" height="{}" style="fill: {}" />
{}
</svg>""".format(
            self.plv["plot_id"],
            self.plv["plot_width"],
            self.plv["Y"],
            "\n  ".join(self.plv["styles"]),
            self.plv["plot_width"],
            self.plv["Y"],
            self.plv["plot_canvas_color"],
            "\n".join(self.plv["pls"])
        )

        return svg

################################################################################
################################################################################
################################################################################



#!/usr/bin/env python3

import cgi
import re, json, yaml
from os import path, environ, pardir
import sys, datetime, argparse
from pymongo import MongoClient

from bycon import *

"""podmd

* https://progenetix.org/services/intervalFrequencies/?datasetIds=progenetix&filters=NCIT:C7376,PMID:22824167
* https://progenetix.org/services/intervalFrequencies/?datasetIds=progenetix&id=pgxcohort-TCGAcancers
* https://progenetix.org/cgi/bycon/services/intervalFrequencies.py/?output=pgxseg&datasetIds=progenetix&filters=NCIT:C7376

podmd"""

################################################################################
################################################################################
################################################################################

def main():

    try:
        interval_frequencies()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)

################################################################################

def intervalFrequencies():
    
    try:
        interval_frequencies()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)
   
################################################################################

def interval_frequencies():

    initialize_bycon_service(byc)

    select_dataset_ids(byc)
    parse_filters(byc)
    parse_variant_parameters(byc)
    generate_genomic_intervals(byc)
    create_empty_service_response(byc)
    cgi_break_on_errors(byc)

    id_rest = rest_path_value("intervalFrequencies")

    if id_rest is not None:
        byc[ "filters" ] = [ {"id": id_rest } ]
    elif "id" in byc["form_data"]:
        byc[ "filters" ] = [ {"id": byc["form_data"]["id"]} ]

    if not "filters" in byc:
        response_add_error(byc, 422, "No value was provided for collation `id` or `filters`.")  
    cgi_break_on_errors(byc)

    # data retrieval & response population
    f_coll_name = byc["config"]["frequencymaps_coll"]
    c_coll_name = byc["config"]["collations_coll"]

    fmap_name = "frequencymap"
    if "codematches" in byc["method"]:
        fmap_name = "frequencymap_codematches"

    results = [ ]

    mongo_client = MongoClient( )
    for ds_id in byc[ "dataset_ids" ]:

        for f in byc[ "filters" ]:

            f_val = f["id"]
 
            collation_f = mongo_client[ ds_id ][ f_coll_name ].find_one( { "id": f_val } )
            collation_c = mongo_client[ ds_id ][ c_coll_name ].find_one( { "id": f_val } )

            if collation_f is None:
                continue

            if "with_samples" in byc["form_data"]: 
                if int(byc["form_data"]["with_samples"]) > 0:
                    if int(collation_c[ "code_matches" ]) < 1:
                        continue

            if not fmap_name in collation_f:
                continue

            if not collation_f:
                response_add_error(byc, 422, "No collation {} was found in {}.{}".format(f_val, ds_id, f_coll_name))
            if not collation_c:
                response_add_error(byc, 422, "No collation {} was found in {}.{}".format(f_val, ds_id, c_coll_name))
            cgi_break_on_errors(byc)

            s_c = collation_c["count"]
            if "analysis_count" in collation_f[ fmap_name ]:
               s_c = collation_f[ fmap_name ]["analysis_count"]

            i_d = collation_c["id"]
            r_o = {
                "dataset_id": ds_id,
                "group_id": i_d,
                "label": re.sub(r';', ',', collation_c["label"]),
                "sample_count": s_c,
                "interval_frequencies": collation_f[ fmap_name ]["intervals"] }
                
            results.append(r_o)

    mongo_client.close( )

    plot_data_bundle = { "interval_frequencies_bundles": results }
    ByconPlot(byc, plot_data_bundle).svg_response()

    check_pgxseg_frequencies_export(byc, results)
    check_pgxmatrix_frequencies_export(byc, results)
    populate_service_response( byc, results)
    cgi_print_response( byc, 200 )

################################################################################

def check_pgxseg_frequencies_export(byc, results):

    if not "pgxseg" in byc["output"] and not "pgxfreq" in byc["output"]:
        return

    export_pgxseg_frequencies(byc, results)

################################################################################

def check_pgxmatrix_frequencies_export(byc, results):

    if not "pgxmatrix" in byc["output"]:
        return

    export_pgxmatrix_frequencies(byc, results)

################################################################################
################################################################################

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import cgi
import re, json, yaml
from os import environ, pardir, path
import sys, os, datetime

from bycon import *

"""podmd

* <https://progenetix.org/beacon/cohorts/>

podmd"""

################################################################################
################################################################################
################################################################################

def main():

    try:
        datasets()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)

################################################################################

def collections():

    try:
        datasets()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)
    
################################################################################

def datasets():

    initialize_bycon_service(byc)
    r, e = instantiate_response_and_error(byc, "beaconCollectionsResponse")
    response_meta_set_info_defaults(r, byc)
    
    _get_history_depth(byc)
    dbstats = datasets_update_latest_stats(byc)

    initialize_beacon_queries(byc)

    if "beaconResultsetsResponse" in byc["response_entity"]["response_schema"]:
        create_empty_beacon_response(byc)
        byc["queries"].pop("datasets", None)
        run_result_sets_beacon(byc)
        query_results_save_handovers(byc)
    else:
        create_empty_beacon_response(byc)
        populate_service_response( byc, dbstats )

        byc["service_response"]["response"]["collections"] = byc["service_response"]["response"].pop("results", None)
        byc["service_response"]["response"].pop("result_sets", None)
        for i, d_s in enumerate(byc["service_response"]["response"]["collections"]):
            # TODO: remove verifier hack
            for t in ["createDateTime", "updateDateTime"]:
                d = str(d_s[t])
                try:
                    if re.match(r'^\d\d\d\d\-\d\d\-\d\d$', d):
                        byc["service_response"]["response"]["collections"][i].update({t:d+"T00:00:00+00:00"})
                except:
                    pass

    cgi_print_response( byc, 200 )

################################################################################

def _get_history_depth(byc):

    if "statsNumber" in byc["form_data"]:
        s_n = byc["form_data"]["statsNumber"]
        try:
            s_n = int(s_n)
        except:
            pass
        if type(s_n) == int:
            if s_n > 0:
                byc.update({"stats_number": s_n})

################################################################################
################################################################################

if __name__ == '__main__':
    main()
